from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.constants import RelationshipStatus
from infrahub.core.graph.schema import GraphAttributeRelationships
from infrahub.core.query import Query
from infrahub.core.schema.generic_schema import GenericSchema

if TYPE_CHECKING:
    from pydantic.fields import FieldInfo

    from infrahub.database import InfrahubDatabase


class AttributeRemoveQuery(Query):
    name = "attribute_remove"
    insert_return: bool = False

    def __init__(
        self,
        attribute_name: str,
        node_kinds: list[str],
        **kwargs: Any,
    ) -> None:
        self.attribute_name = attribute_name
        self.node_kinds = node_kinds
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string())
        self.params.update(branch_params)

        kinds_to_ignore = []
        profile_kinds_to_update = []

        for node_kind in self.node_kinds:
            new_schema = db.schema.get(name=node_kind, branch=self.branch, duplicate=False)

        if isinstance(new_schema, GenericSchema):
            for inheriting_schema_kind in new_schema.used_by:
                node_schema = db.schema.get_node_schema(
                    name=inheriting_schema_kind, branch=self.branch, duplicate=False
                )
                attr_schema = node_schema.get_attribute_or_none(name=self.attribute_name)
                if attr_schema and not attr_schema.inherited:
                    kinds_to_ignore.append(inheriting_schema_kind)
                else:
                    profile_kinds_to_update.append(f"Profile{inheriting_schema_kind}")

        self.params["kinds_to_ignore"] = kinds_to_ignore
        self.params["attr_name"] = self.attribute_name
        self.params["current_time"] = self.at.to_string()
        self.params["branch_name"] = self.branch.name

        self.params["rel_props"] = {
            "branch": self.branch.name,
            "branch_level": self.branch.hierarchy_level,
            "status": RelationshipStatus.DELETED.value,
            "from": self.at.to_string(),
        }

        def render_sub_query_per_rel_type(rel_type: str, rel_def: FieldInfo) -> str:
            subquery = [
                "WITH peer_node, rb, active_attr",
                f'WHERE type(rb) = "{rel_type}"',
            ]
            if rel_def.default.direction.value == "outbound":
                subquery.append(f"CREATE (active_attr)-[:{rel_type} $rel_props ]->(peer_node)")
            elif rel_def.default.direction.value == "inbound":
                subquery.append(f"CREATE (active_attr)<-[:{rel_type} $rel_props ]-(peer_node)")
            else:
                subquery.append(f"CREATE (active_attr)-[:{rel_type} $rel_props ]-(peer_node)")

            subquery.append("RETURN peer_node as p2")
            return "\n".join(subquery)

        sub_queries = [
            render_sub_query_per_rel_type(rel_type, rel_def)
            for rel_type, rel_def in GraphAttributeRelationships.model_fields.items()
        ]
        sub_query_all = "\nUNION\n".join(sub_queries)

        node_kinds_str = "|".join(self.node_kinds + profile_kinds_to_update)
        query = """
        // Find all the active nodes
        MATCH (node:%(node_kinds)s)
        WHERE (size($kinds_to_ignore) = 0 OR NOT any(l IN labels(node) WHERE l IN $kinds_to_ignore))
        AND exists((node)-[:HAS_ATTRIBUTE]-(:Attribute { name: $attr_name }))
        CALL (node) {
            MATCH (root:Root)<-[r:IS_PART_OF]-(node)
            WHERE %(branch_filter)s
            RETURN node as n1, r as r1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as active_node, r1 as rb
        WHERE rb.status = "active"
        // Find all the attributes that need to be updated
        CALL (active_node) {
            MATCH (active_node)-[r:HAS_ATTRIBUTE]-(attr:Attribute { name: $attr_name })
            WHERE %(branch_filter)s
            RETURN active_node as n1, r as r1, attr as attr1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as active_node, r1 as rb, attr1 as active_attr
        WHERE rb.status = "active"
        WITH active_attr
        MATCH (active_attr)-[]-(peer)
        WITH DISTINCT active_attr, peer
        CALL (active_attr, peer) {
            MATCH (active_attr)-[r]-(peer)
            WHERE %(branch_filter)s
            RETURN active_attr as a1, r as r1, peer as p1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH a1 as active_attr, r1 as rb, p1 as peer_node
        WHERE rb.status = "active"
        CALL (peer_node, rb, active_attr) {
            %(sub_query_all)s
        }
        WITH p2 as peer_node, rb, active_attr
        FOREACH (i in CASE WHEN rb.branch = $branch_name THEN [1] ELSE [] END |
            SET rb.to = $current_time
        )
        RETURN DISTINCT active_attr
        """ % {
            "branch_filter": branch_filter,
            "sub_query_all": sub_query_all,
            "node_kinds": node_kinds_str,
        }
        self.add_to_query(query)



from __future__ import annotations

from prefect import flow
from prefect.logging import get_run_logger

from infrahub.workers.dependencies import get_client, get_workflow
from infrahub.workflows.catalogue import PROFILE_REFRESH
from infrahub.workflows.utils import add_tags

REFRESH_PROFILES_MUTATION = """
mutation RefreshProfiles(
    $id: String!,
  ) {
  InfrahubProfilesRefresh(
    data: {id: $id}
  ) {
    ok
  }
}
"""


@flow(
    name="object-profiles-refresh",
    flow_run_name="Refresh profiles for {node_id}",
)
async def object_profiles_refresh(
    branch_name: str,
    node_id: str,
) -> None:
    log = get_run_logger()
    client = get_client()

    await add_tags(branches=[branch_name], nodes=[node_id], db_change=True)
    await client.execute_graphql(
        query=REFRESH_PROFILES_MUTATION,
        variables={"id": node_id},
        branch_name=branch_name,
    )
    log.info(f"Profiles refreshed for {node_id}")


@flow(
    name="objects-profiles-refresh-multiple",
    flow_run_name="Refresh profiles for multiple objects",
)
async def objects_profiles_refresh_multiple(
    branch_name: str,
    node_ids: list[str],
) -> None:
    log = get_run_logger()

    await add_tags(branches=[branch_name])

    for node_id in node_ids:
        log.info(f"Requesting profile refresh for {node_id}")
        await get_workflow().submit_workflow(
            workflow=PROFILE_REFRESH,
            parameters={
                "branch_name": branch_name,
                "node_id": node_id,
            },
        )



from __future__ import annotations

from pathlib import Path

import yaml

from ..schema.repository import InfrahubRepositoryConfig
from .exceptions import FileNotValidError


def find_repository_config_file(base_path: Path | None = None) -> Path:
    """Find the repository config file, checking for both .yml and .yaml extensions.

    Args:
        base_path: Base directory to search in. If None, uses current directory.

    Returns:
        Path to the config file.

    Raises:
        FileNotFoundError: If neither .infrahub.yml nor .infrahub.yaml exists.
    """
    if base_path is None:
        base_path = Path()

    yml_path = base_path / ".infrahub.yml"
    yaml_path = base_path / ".infrahub.yaml"

    # Prefer .yml if both exist
    if yml_path.exists():
        return yml_path
    if yaml_path.exists():
        return yaml_path
    # For backward compatibility, return .yml path for error messages
    return yml_path


def load_repository_config(repo_config_file: Path) -> InfrahubRepositoryConfig:
    # If the file doesn't exist, try to find it with alternate extension
    if not repo_config_file.exists():
        if repo_config_file.name == ".infrahub.yml":
            alt_path = repo_config_file.parent / ".infrahub.yaml"
            if alt_path.exists():
                repo_config_file = alt_path
        elif repo_config_file.name == ".infrahub.yaml":
            alt_path = repo_config_file.parent / ".infrahub.yml"
            if alt_path.exists():
                repo_config_file = alt_path

    if not repo_config_file.is_file():
        raise FileNotFoundError(repo_config_file)

    try:
        yaml_data = repo_config_file.read_text()
        data = yaml.safe_load(yaml_data)
    except yaml.YAMLError as exc:
        raise FileNotValidError(name=str(repo_config_file)) from exc

    return InfrahubRepositoryConfig(**data)



# Generated by "invoke backend.generate", do not edit directly

from __future__ import annotations

from typing import Any

from pydantic import Field

from infrahub.core.constants import AllowOverrideType, BranchSupportType, HashableModelState
from infrahub.core.models import HashableModel
from infrahub.core.schema.attribute_parameters import (
    AttributeParameters,
    NumberAttributeParameters,
    NumberPoolParameters,
    TextAttributeParameters,
)
from infrahub.core.schema.computed_attribute import ComputedAttribute  # noqa: TC001
from infrahub.core.schema.dropdown import DropdownChoice  # noqa: TC001


class GeneratedAttributeSchema(HashableModel):
    id: str | None = Field(
        default=None, description="The ID of the attribute", json_schema_extra={"update": "not_applicable"}
    )
    name: str = Field(
        ...,
        description="Attribute name, must be unique within a model and must be all lowercase.",
        pattern=r"^[a-z0-9\_]+$",
        min_length=3,
        max_length=32,
        json_schema_extra={"update": "migration_required"},
    )
    kind: str = Field(
        ..., description="Defines the type of the attribute.", json_schema_extra={"update": "migration_required"}
    )
    enum: list | None = Field(
        default=None,
        description="Define a list of valid values for the attribute.",
        json_schema_extra={"update": "validate_constraint"},
    )
    computed_attribute: ComputedAttribute | None = Field(
        default=None,
        description="Defines how the value of this attribute will be populated.",
        json_schema_extra={"update": "allowed"},
    )
    choices: list[DropdownChoice] | None = Field(
        default=None,
        description="Define a list of valid choices for a dropdown attribute.",
        json_schema_extra={"update": "validate_constraint"},
    )
    regex: str | None = Field(
        default=None,
        description="Regex uses to limit the characters allowed in for the attributes. (deprecated: please use parameters.regex instead)",
        json_schema_extra={"update": "validate_constraint"},
    )
    max_length: int | None = Field(
        default=None,
        description="Set a maximum number of characters allowed for a given attribute. (deprecated: please use parameters.max_length instead)",
        json_schema_extra={"update": "validate_constraint"},
    )
    min_length: int | None = Field(
        default=None,
        description="Set a minimum number of characters allowed for a given attribute. (deprecated: please use parameters.min_length instead)",
        json_schema_extra={"update": "validate_constraint"},
    )
    label: str | None = Field(
        default=None,
        description="Human friendly representation of the name. Will be autogenerated if not provided",
        max_length=32,
        json_schema_extra={"update": "allowed"},
    )
    description: str | None = Field(
        default=None,
        description="Short description of the attribute.",
        max_length=128,
        json_schema_extra={"update": "allowed"},
    )
    read_only: bool = Field(
        default=False,
        description="Set the attribute as Read-Only, users won't be able to change its value. Mainly relevant for internal object.",
        json_schema_extra={"update": "migration_required"},
    )
    unique: bool = Field(
        default=False,
        description="Indicate if the value of this attribute must be unique in the database for a given model.",
        json_schema_extra={"update": "validate_constraint"},
    )
    optional: bool = Field(
        default=False,
        description="Indicate if this attribute is mandatory or optional.",
        json_schema_extra={"update": "migration_required"},
    )
    branch: BranchSupportType | None = Field(
        default=None,
        description="Type of branch support for the attribute, if not defined it will be inherited from the node.",
        json_schema_extra={"update": "not_supported"},
    )
    order_weight: int | None = Field(
        default=None,
        description="Number used to order the attribute in the frontend (table and view). Lowest value will be ordered first.",
        json_schema_extra={"update": "allowed"},
    )
    default_value: Any | None = Field(
        default=None, description="Default value of the attribute.", json_schema_extra={"update": "allowed"}
    )
    inherited: bool = Field(
        default=False,
        description="Internal value to indicate if the attribute was inherited from a Generic node.",
        json_schema_extra={"update": "not_applicable"},
    )
    state: HashableModelState = Field(
        default=HashableModelState.PRESENT,
        description="Expected state of the attribute after loading the schema",
        json_schema_extra={"update": "not_applicable"},
    )
    allow_override: AllowOverrideType = Field(
        default=AllowOverrideType.ANY,
        description="Type of allowed override for the attribute.",
        json_schema_extra={"update": "allowed"},
    )
    parameters: AttributeParameters | TextAttributeParameters | NumberAttributeParameters | NumberPoolParameters = (
        Field(
            default_factory=AttributeParameters,
            description="Extra parameters specific to this kind of attribute",
            json_schema_extra={"update": "validate_constraint"},
        )
    )
    deprecation: str | None = Field(
        default=None,
        description="Mark attribute as deprecated and provide a user-friendly message to display",
        max_length=128,
        json_schema_extra={"update": "allowed"},
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from infrahub.workers.dependencies import get_database

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


@dataclass
class CliContext:
    application: str = "infrahub.server:app"

    # This method is inherited for Infrahub Enterprise.
    @staticmethod
    async def init_db(retry: int) -> InfrahubDatabase:  # type:ignore  # noqa
        return await get_database()



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.query import QueryType

from .standard_node import StandardNodeQuery

if TYPE_CHECKING:
    from infrahub.core.task import TaskLog
    from infrahub.database import InfrahubDatabase


class TaskLogNodeCreateQuery(StandardNodeQuery):
    name: str = "log_create"
    node: TaskLog

    type: QueryType = QueryType.WRITE

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        node_type = self.node.get_type()
        self.params["node_prop"] = self.node.to_db()
        self.params["task_id"] = self.node.task_id

        # If the request to create the Log arrives earlier that that of the creation of the Task
        # we want the relationship to be setup regardless so that it's in place for when the
        # Task gets propperly created.
        query = """
        MATCH (t:Task { uuid: $task_id })
        CREATE (n:%(node_type)s $node_prop)-[:RELATES_TO]->(t)
        """ % {"node_type": node_type}
        self.add_to_query(query=query)
        self.return_labels = ["n"]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub import config
from infrahub.core.constants import GLOBAL_BRANCH_NAME
from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class DeleteBranchRelationshipsQuery(Query):
    name: str = "delete_branch_relationships"
    insert_return: bool = False

    type: QueryType = QueryType.WRITE

    def __init__(self, branch_name: str, **kwargs: Any):
        self.branch_name = branch_name
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        query = """
// --------------
// for every Node created on this branch (it's about to be deleted), find any agnostic relationships
// connected to the Node and delete them
// --------------
OPTIONAL MATCH (:Root)<-[e:IS_PART_OF {status: "active"}]-(n:Node)
WHERE e.branch = $branch_name
CALL (n) {
    OPTIONAL MATCH (n)-[:IS_RELATED {branch: $global_branch_name}]-(rel:Relationship)
    DETACH DELETE rel
} IN TRANSACTIONS OF 500 ROWS

// reduce the results to a single row
WITH 1 AS one
LIMIT 1

// --------------
// for every edge on this branch, delete it
// --------------
MATCH (s)-[r]->(d)
WHERE r.branch = $branch_name
CALL (r) {
    DELETE r
} IN TRANSACTIONS OF 500 ROWS

// --------------
// get the database IDs of every vertex linked to a deleted edge
// --------------
WITH DISTINCT elementId(s) AS s_id, elementId(d) AS d_id
WITH collect(s_id) + collect(d_id) AS vertex_ids
UNWIND vertex_ids AS vertex_id

// --------------
// delete any vertices that are now orphaned
// --------------
CALL (vertex_id) {
    MATCH (n)
    WHERE elementId(n) = vertex_id
    AND NOT exists((n)--())
    DELETE n
} IN TRANSACTIONS OF 500 ROWS
        """
        self.params["branch_name"] = self.branch_name
        self.params["global_branch_name"] = GLOBAL_BRANCH_NAME
        self.add_to_query(query)


class GetAllBranchInternalRelationshipQuery(Query):
    name: str = "get_internal_relationship"

    type: QueryType = QueryType.READ
    insert_return: bool = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        query = """
        MATCH p = ()-[r]-()
        WHERE r.branch = $branch_name
        RETURN DISTINCT r
        """
        self.add_to_query(query=query)
        self.params["branch_name"] = self.branch.name
        self.return_labels = ["r"]


class RebaseBranchUpdateRelationshipQuery(Query):
    name: str = "rebase_branch_update"

    type: QueryType = QueryType.WRITE

    def __init__(self, ids: list[str], **kwargs: Any) -> None:
        self.ids = ids
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        query = """
        MATCH ()-[r]->()
        WHERE %(id_func)s(r) IN $ids
        SET r.from = $at
        SET r.conflict = NULL
        """ % {
            "id_func": db.get_id_function_name(),
        }

        self.add_to_query(query=query)

        self.params["at"] = self.at.to_string()
        self.params["ids"] = [db.to_database_id(id) for id in self.ids]
        self.return_labels = [f"{db.get_id_function_name()}(r)"]


class RebaseBranchDeleteRelationshipQuery(Query):
    name: str = "rebase_branch_delete"

    type: QueryType = QueryType.WRITE
    insert_return: bool = False

    def __init__(self, ids: list[str], **kwargs: Any) -> None:
        self.ids = ids
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        if config.SETTINGS.database.db_type == config.DatabaseType.MEMGRAPH:
            query = """
            MATCH p = (s)-[r]-(d)
            WHERE %(id_func)s(r) IN $ids
            DELETE r
            """
        else:
            query = """
            MATCH p = (s)-[r]-(d)
            WHERE %(id_func)s(r) IN $ids
            DELETE r
            WITH *
            UNWIND nodes(p) AS n
            MATCH (n)
            WHERE NOT exists((n)--())
            DELETE n
            """
        query %= {
            "id_func": db.get_id_function_name(),
        }

        self.add_to_query(query=query)

        self.params["ids"] = [db.to_database_id(id) for id in self.ids]



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core.constants import AllowOverrideType, InfrahubKind, RelationshipKind

from .generated.node_schema import GeneratedNodeSchema
from .generic_schema import GenericSchema

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase


class NodeSchema(GeneratedNodeSchema):
    @property
    def is_node_schema(self) -> bool:
        return True

    @property
    def is_generic_schema(self) -> bool:
        return False

    @property
    def is_profile_schema(self) -> bool:
        return False

    @property
    def is_template_schema(self) -> bool:
        return False

    @property
    def is_ip_prefix(self) -> bool:
        """Return whether a node is a derivative of built-in IP prefixes."""
        return InfrahubKind.IPPREFIX in self.inherit_from

    @property
    def is_ip_address(self) -> bool:
        """Return whether a node is a derivative of built-in IP addreses."""
        return InfrahubKind.IPADDRESS in self.inherit_from

    def validate_inheritance(self, interface: GenericSchema) -> None:
        """Perform checks specific to inheritance from Generics.

        Checks:
            - Check that protected attributes and relationships are not overridden before inheriting them from interface.
            - Check that the attribute types to be inherited are same kind.
        """
        for attribute in self.attributes:
            if attribute.name in interface.attribute_names:
                if (
                    not attribute.inherited
                    and interface.get_attribute(attribute.name).allow_override == AllowOverrideType.NONE
                ):
                    raise ValueError(
                        f"{self.kind}'s attribute {attribute.name} inherited from {interface.kind} cannot be overriden"
                    )

                interface_attr = interface.get_attribute(attribute.name)
                # Check existing inherited attribute kind is the same as the incoming inherited attribute
                if attribute.kind != interface_attr.kind:
                    raise ValueError(
                        f"{self.kind}.{attribute.name} inherited from {interface.namespace}{interface.name} must be the same kind "
                        f'["{interface_attr.kind}", "{attribute.kind}"]'
                    )

        for relationship in self.relationships:
            if relationship.name in interface.relationship_names and not relationship.inherited:
                interface_relationship = interface.get_relationship(relationship.name)
                if interface_relationship.allow_override == AllowOverrideType.NONE:
                    raise ValueError(
                        f"{self.kind}'s relationship {relationship.name} inherited from {interface.kind} cannot be overriden"
                    )
                if relationship.kind != RelationshipKind.HIERARCHY and relationship.peer != interface_relationship.peer:
                    raise ValueError(
                        f"{self.kind}'s relationship {relationship.name} inherited from {interface.kind} must have the same peer "
                        f"({interface_relationship.peer} != {relationship.peer})"
                    )

    def inherit_from_interface(self, interface: GenericSchema) -> None:
        existing_inherited_attributes: dict[str, int] = {
            item.name: idx for idx, item in enumerate(self.attributes) if item.inherited
        }
        existing_inherited_relationships: dict[str, int] = {
            item.name: idx for idx, item in enumerate(self.relationships) if item.inherited
        }
        existing_inherited_fields = list(existing_inherited_attributes.keys()) + list(
            existing_inherited_relationships.keys()
        )

        properties_to_inherit = [
            "human_friendly_id",
            "display_label",
            "display_labels",
            "default_filter",
            "menu_placement",
            "uniqueness_constraints",
            "icon",
            "order_by",
        ]
        for prop_name in properties_to_inherit:
            if getattr(interface, prop_name) and not getattr(self, prop_name):
                setattr(self, prop_name, getattr(interface, prop_name))

        for attribute in interface.attributes:
            if attribute.name in self.valid_local_names:
                continue

            new_attribute = attribute.duplicate()
            new_attribute.id = None
            new_attribute.inherited = True

            if attribute.name not in existing_inherited_fields:
                self.attributes.append(new_attribute)
            else:
                item_idx = existing_inherited_attributes[attribute.name]
                self.attributes[item_idx].update_from_generic(other=new_attribute)

        for relationship in interface.relationships:
            if relationship.name in self.valid_local_names:
                continue

            new_relationship = relationship.duplicate()
            new_relationship.id = None
            new_relationship.inherited = True

            if relationship.name not in existing_inherited_fields:
                self.relationships.append(new_relationship)
            else:
                item_idx = existing_inherited_relationships[relationship.name]
                self.relationships[item_idx].update_from_generic(other=new_relationship)

    def get_hierarchy_schema(
        self, db: InfrahubDatabase, branch: Branch | str | None = None, duplicate: bool = False
    ) -> GenericSchema:
        if not self.hierarchy:
            raise ValueError("The node is not part of a hierarchy")
        schema = db.schema.get(name=self.hierarchy, branch=branch, duplicate=duplicate)
        if not isinstance(schema, GenericSchema):
            raise TypeError
        return schema

    def get_labels(self) -> list[str]:
        """Return the labels for this object, composed of the kind
        and the list of Generic this object is inheriting from."""

        labels: list[str] = [self.kind] + self.inherit_from
        if self.namespace not in ["Schema", "Internal"] and InfrahubKind.GENERICGROUP not in self.inherit_from:
            labels.append(InfrahubKind.NODE)
        return labels



from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.schema import SchemaRoot, internal_schema
from infrahub.core.schema.manager import SchemaManager
from infrahub.core.schema.schema_branch import SchemaBranch
from infrahub.database import InfrahubDatabase
from infrahub.exceptions import InitializationError


async def get_or_load_schema_branch(db: InfrahubDatabase, branch: Branch) -> SchemaBranch:
    try:
        if registry.schema.has_schema_branch(branch.name):
            return registry.schema.get_schema_branch(branch.name)
    except InitializationError:
        schema_manager = SchemaManager()
        registry.schema = schema_manager
        internal_schema_root = SchemaRoot(**internal_schema)
        registry.schema.register_schema(schema=internal_schema_root)
    schema_branch = await registry.schema.load_schema_from_db(db=db, branch=branch)
    registry.schema.set_schema_branch(name=branch.name, schema=schema_branch)
    return schema_branch



from typing import Any

from infrahub.core.attribute import BaseAttribute
from infrahub.core.branch import Branch
from infrahub.core.node import Node
from infrahub.database import InfrahubDatabase

from .queries.get_profile_data import GetProfileDataQuery, ProfileData


class NodeProfilesApplier:
    def __init__(self, db: InfrahubDatabase, branch: Branch):
        self.db = db
        self.branch = branch

    async def _get_profile_ids(self, node: Node) -> list[str]:
        try:
            profiles_rel = node.get_relationship("profiles")
        except ValueError:
            return []
        profile_rels = await profiles_rel.get_relationships(db=self.db)
        return [pr.peer_id for pr in profile_rels if pr.peer_id]

    async def _get_attr_names_for_profiles(self, node: Node) -> list[str]:
        node_schema = node.get_schema()

        # get the names of attributes that could be affected by profile changes
        attr_names_for_profiles: list[str] = []
        for attr_schema in node_schema.attributes:
            attr_name = attr_schema.name
            node_attr: BaseAttribute = getattr(node, attr_name)
            if node_attr.is_from_profile or node_attr.is_default:
                attr_names_for_profiles.append(attr_name)
        return attr_names_for_profiles

    async def _get_sorted_profile_data(
        self, profile_ids: list[str], attr_names_for_profiles: list[str]
    ) -> list[ProfileData]:
        if not profile_ids:
            return []
        query = await GetProfileDataQuery.init(
            db=self.db, branch=self.branch, profile_ids=profile_ids, attr_names=attr_names_for_profiles
        )
        await query.execute(db=self.db)
        profile_data_list = query.get_profile_data()
        return sorted(profile_data_list, key=lambda x: (x.priority, x.uuid))

    def _apply_profile_to_attribute(self, node_attr: BaseAttribute, profile_value: Any, profile_id: str) -> bool:
        is_changed = False
        if node_attr.value != profile_value:
            node_attr.value = profile_value
            is_changed = True
        if node_attr.is_default is not False:
            node_attr.is_default = False
            is_changed = True
        if node_attr.is_from_profile is not True:
            node_attr.is_from_profile = True
            is_changed = True
        if node_attr.source_id != profile_id:  # type: ignore[attr-defined]
            node_attr.set_source(value=profile_id)
            is_changed = True
        return is_changed

    def _remove_profile_from_attribute(self, node_attr: BaseAttribute) -> None:
        node_attr.clear_source()
        node_attr.value = node_attr.schema.default_value
        node_attr.is_default = True
        node_attr.is_from_profile = False

    async def apply_profiles(self, node: Node) -> list[str]:
        profile_ids = await self._get_profile_ids(node=node)
        attr_names_for_profiles = await self._get_attr_names_for_profiles(node=node)

        if not attr_names_for_profiles:
            return []

        # get profiles priorities and attribute values on branch
        sorted_profile_data = await self._get_sorted_profile_data(
            profile_ids=profile_ids, attr_names_for_profiles=attr_names_for_profiles
        )

        updated_field_names = []
        # set attribute values/is_default/is_from_profile on nodes
        for attr_name in attr_names_for_profiles:
            has_profile_data = False
            node_attr = node.get_attribute(attr_name)
            for profile_data in sorted_profile_data:
                profile_value = profile_data.attribute_values.get(attr_name)
                if profile_value is not None:
                    has_profile_data = True
                    is_changed = False
                    is_changed = self._apply_profile_to_attribute(
                        node_attr=node_attr, profile_value=profile_value, profile_id=profile_data.uuid
                    )
                    if is_changed:
                        updated_field_names.append(attr_name)
                    break
            if not has_profile_data and node_attr.is_from_profile:
                self._remove_profile_from_attribute(node_attr=node_attr)
                updated_field_names.append(attr_name)
        return updated_field_names



from infrahub.core.constants import (
    AllowOverrideType,
    BranchSupportType,
    InfrahubKind,
    RepositoryInternalStatus,
    RepositoryOperationalStatus,
    RepositorySyncStatus,
)
from infrahub.core.constants import RelationshipCardinality as Cardinality
from infrahub.core.constants import RelationshipKind as RelKind

from ...attribute_schema import AttributeSchema as Attr
from ...dropdown import DropdownChoice
from ...generic_schema import GenericSchema
from ...node_schema import NodeSchema
from ...relationship_schema import (
    RelationshipSchema as Rel,
)

core_repository = NodeSchema(
    name="Repository",
    namespace="Core",
    description="A Git Repository integrated with Infrahub",
    include_in_menu=False,
    icon="mdi:source-repository",
    label="Repository",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    generate_profile=False,
    branch=BranchSupportType.AGNOSTIC,
    inherit_from=[
        InfrahubKind.LINEAGEOWNER,
        InfrahubKind.LINEAGESOURCE,
        InfrahubKind.GENERICREPOSITORY,
        InfrahubKind.TASKTARGET,
    ],
    documentation="/topics/repository",
    attributes=[
        Attr(
            name="default_branch",
            kind="Text",
            default_value="main",
            order_weight=6000,
        ),
        Attr(
            name="commit",
            kind="Text",
            optional=True,
            branch=BranchSupportType.LOCAL,
            order_weight=7000,
        ),
    ],
)

core_read_only_repository = NodeSchema(
    name="ReadOnlyRepository",
    namespace="Core",
    description="A Git Repository integrated with Infrahub, Git-side will not be updated",
    include_in_menu=False,
    label="Read-Only Repository",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    generate_profile=False,
    branch=BranchSupportType.AGNOSTIC,
    inherit_from=[
        InfrahubKind.LINEAGEOWNER,
        InfrahubKind.LINEAGESOURCE,
        InfrahubKind.GENERICREPOSITORY,
        InfrahubKind.TASKTARGET,
    ],
    documentation="/topics/repository",
    attributes=[
        Attr(
            name="ref",
            kind="Text",
            default_value="main",
            branch=BranchSupportType.AWARE,
            order_weight=6000,
        ),
        Attr(
            name="commit",
            kind="Text",
            optional=True,
            branch=BranchSupportType.AWARE,
            order_weight=7000,
        ),
    ],
)

core_generic_repository = GenericSchema(
    name="GenericRepository",
    namespace="Core",
    label="Git Repository",
    description="A Git Repository integrated with Infrahub",
    include_in_menu=False,
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    icon="mdi:source-repository",
    branch=BranchSupportType.AGNOSTIC,
    uniqueness_constraints=[["name__value"], ["location__value"]],
    documentation="/topics/repository",
    attributes=[
        Attr(
            name="name",
            regex=r"^[^/]*$",
            kind="Text",
            unique=True,
            branch=BranchSupportType.AGNOSTIC,
            order_weight=1000,
            allow_override=AllowOverrideType.NONE,
        ),
        Attr(
            name="description",
            kind="Text",
            optional=True,
            branch=BranchSupportType.AGNOSTIC,
            order_weight=2000,
            allow_override=AllowOverrideType.NONE,
        ),
        Attr(
            name="location",
            kind="Text",
            unique=True,
            branch=BranchSupportType.AGNOSTIC,
            order_weight=3000,
            allow_override=AllowOverrideType.NONE,
        ),
        Attr(
            name="internal_status",
            kind="Dropdown",
            choices=[
                DropdownChoice(
                    name=RepositoryInternalStatus.STAGING.value,
                    label="Staging",
                    description="Repository was recently added to this branch.",
                    color="#fef08a",
                ),
                DropdownChoice(
                    name=RepositoryInternalStatus.ACTIVE.value,
                    label="Active",
                    description="Repository is actively being synced for this branch",
                    color="#86efac",
                ),
                DropdownChoice(
                    name=RepositoryInternalStatus.INACTIVE.value,
                    label="Inactive",
                    description="Repository is not active on this branch.",
                    color="#e5e7eb",
                ),
            ],
            default_value="inactive",
            optional=False,
            branch=BranchSupportType.LOCAL,
            order_weight=7000,
            allow_override=AllowOverrideType.NONE,
        ),
        Attr(
            name="operational_status",
            kind="Dropdown",
            choices=[
                DropdownChoice(
                    name=RepositoryOperationalStatus.UNKNOWN.value,
                    label="Unknown",
                    description="Status of the repository is unknown and mostlikely because it hasn't been synced yet",
                    color="#9ca3af",
                ),
                DropdownChoice(
                    name=RepositoryOperationalStatus.ONLINE.value,
                    label="Online",
                    description="Repository connection is working",
                    color="#86efac",
                ),
                DropdownChoice(
                    name=RepositoryOperationalStatus.ERROR_CRED.value,
                    label="Credential Error",
                    description="Repository can't be synced due to some credential error(s)",
                    color="#f87171",
                ),
                DropdownChoice(
                    name=RepositoryOperationalStatus.ERROR_CONNECTION.value,
                    label="Connectivity Error",
                    description="Repository can't be synced due to some connectivity error(s)",
                    color="#f87171",
                ),
                DropdownChoice(
                    name=RepositoryOperationalStatus.ERROR.value,
                    label="Error",
                    description="Repository can't be synced due to an unknown error",
                    color="#ef4444",
                ),
            ],
            optional=False,
            branch=BranchSupportType.AGNOSTIC,
            default_value=RepositoryOperationalStatus.UNKNOWN.value,
            order_weight=5000,
        ),
        Attr(
            name="sync_status",
            kind="Dropdown",
            choices=[
                DropdownChoice(
                    name=RepositorySyncStatus.UNKNOWN.value,
                    label="Unknown",
                    description="Status of the repository is unknown and mostlikely because it hasn't been synced yet",
                    color="#9ca3af",
                ),
                DropdownChoice(
                    name=RepositorySyncStatus.ERROR_IMPORT.value,
                    label="Import Error",
                    description="Repository import error observed",
                    color="#f87171",
                ),
                DropdownChoice(
                    name=RepositorySyncStatus.IN_SYNC.value,
                    label="In Sync",
                    description="The repository is syncing correctly",
                    color="#60a5fa",
                ),
                DropdownChoice(
                    name=RepositorySyncStatus.SYNCING.value,
                    label="Syncing",
                    description="A sync job is currently running against the repository.",
                    color="#a855f7",
                ),
            ],
            optional=False,
            branch=BranchSupportType.LOCAL,
            default_value=RepositorySyncStatus.UNKNOWN.value,
            order_weight=6000,
        ),
    ],
    relationships=[
        Rel(
            name="credential",
            peer=InfrahubKind.CREDENTIAL,
            identifier="gitrepository__credential",
            kind=RelKind.ATTRIBUTE,
            optional=True,
            cardinality=Cardinality.ONE,
            order_weight=4000,
        ),
        Rel(
            name="tags",
            peer=InfrahubKind.TAG,
            kind=RelKind.ATTRIBUTE,
            optional=True,
            cardinality=Cardinality.MANY,
            order_weight=8000,
        ),
        Rel(
            name="transformations",
            peer=InfrahubKind.TRANSFORM,
            identifier="repository__transformation",
            optional=True,
            cardinality=Cardinality.MANY,
            order_weight=10000,
        ),
        Rel(
            name="queries",
            peer=InfrahubKind.GRAPHQLQUERY,
            identifier="graphql_query__repository",
            optional=True,
            cardinality=Cardinality.MANY,
            order_weight=9000,
        ),
        Rel(
            name="checks",
            peer=InfrahubKind.CHECKDEFINITION,
            identifier="check_definition__repository",
            optional=True,
            cardinality=Cardinality.MANY,
            order_weight=11000,
        ),
        Rel(
            name="generators",
            peer=InfrahubKind.GENERATORDEFINITION,
            identifier="generator_definition__repository",
            optional=True,
            cardinality=Cardinality.MANY,
            order_weight=12000,
        ),
        Rel(
            name="groups_objects",
            peer=InfrahubKind.REPOSITORYGROUP,
            optional=True,
            cardinality=Cardinality.MANY,
            order_weight=13000,
        ),
    ],
)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import GraphMigration, MigrationResult
from infrahub.log import get_logger

from ...query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class DeletePosthumousEdges(Query):
    name = "delete_posthumous_edges_query"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
// ------------
// find deleted nodes
// ------------
MATCH (n:Node)-[e:IS_PART_OF]->(:Root)
WHERE e.status = "deleted" OR e.to IS NOT NULL
WITH DISTINCT n, e.branch AS delete_branch, e.branch_level AS delete_branch_level, CASE
    WHEN e.status = "deleted" THEN e.from
    ELSE e.to
END AS delete_time
// ------------
// find the edges added to the deleted node after the delete time
// ------------
MATCH (n)-[added_e]-(peer)
WHERE added_e.from > delete_time
AND type(added_e) <> "IS_PART_OF"
// if the node was deleted on a branch (delete_branch_level > 1), and then updated on main/global (added_e.branch_level = 1), we can ignore it
AND added_e.branch_level >= delete_branch_level
AND (added_e.branch = delete_branch OR delete_branch_level = 1)
WITH DISTINCT n, delete_branch, delete_time, added_e, peer
// ------------
// get the branched_from for the branch on which the node was deleted
// ------------
CALL (added_e) {
    MATCH (b:Branch {name: added_e.branch})
    RETURN b.branched_from AS added_e_branched_from
}
// ------------
// account for the following situations, given that the edge update time is after the node delete time
//  - deleted on main/global, updated on branch
//    - illegal if the delete is before branch.branched_from
//  - deleted on branch, updated on branch
//    - illegal
// ------------
WITH n, delete_branch, delete_time, added_e, peer
WHERE delete_branch = added_e.branch
OR delete_time < added_e_branched_from
DELETE added_e
// --------------
// the peer _should_ only be an Attribute, but I want to make sure we don't
// inadvertently delete Root or an AttributeValue or a Boolean
// --------------
WITH peer
WHERE "Attribute" IN labels(peer)
DETACH DELETE peer
        """
        self.add_to_query(query)


class Migration030(GraphMigration):
    """
    Edges could have been added to Nodes after the Node was deleted, so we need to hard-delete those illegal edges
    """

    name: str = "030_delete_illegal_edges"
    minimum_version: int = 29
    queries: Sequence[type[Query]] = [DeletePosthumousEdges]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        result = MigrationResult()
        return result



from __future__ import annotations

import linecache
from pathlib import Path
from typing import Any, Callable, NoReturn

import jinja2
from jinja2 import meta, nodes
from jinja2.sandbox import SandboxedEnvironment
from netutils.utils import jinja2_convenience_function
from rich.syntax import Syntax
from rich.traceback import Traceback

from .exceptions import (
    JinjaTemplateError,
    JinjaTemplateNotFoundError,
    JinjaTemplateOperationViolationError,
    JinjaTemplateSyntaxError,
    JinjaTemplateUndefinedError,
)
from .filters import AVAILABLE_FILTERS
from .models import UndefinedJinja2Error

netutils_filters = jinja2_convenience_function()


class Jinja2Template:
    def __init__(
        self,
        template: str | Path,
        template_directory: Path | None = None,
        filters: dict[str, Callable] | None = None,
    ) -> None:
        self.is_string_based = isinstance(template, str)
        self.is_file_based = isinstance(template, Path)
        self._template = str(template)
        self._template_directory = template_directory
        self._environment: jinja2.Environment | None = None

        self._available_filters = [filter_definition.name for filter_definition in AVAILABLE_FILTERS]
        self._trusted_filters = [
            filter_definition.name for filter_definition in AVAILABLE_FILTERS if filter_definition.trusted
        ]

        self._filters = filters or {}
        for user_filter in self._filters:
            self._available_filters.append(user_filter)
            self._trusted_filters.append(user_filter)

        self._template_definition: jinja2.Template | None = None

    def get_environment(self) -> jinja2.Environment:
        if self._environment:
            return self._environment

        if self.is_string_based:
            return self._get_string_based_environment()

        return self._get_file_based_environment()

    def get_template(self) -> jinja2.Template:
        if self._template_definition:
            return self._template_definition

        try:
            if self.is_string_based:
                template = self._get_string_based_template()
            else:
                template = self._get_file_based_template()
        except jinja2.TemplateSyntaxError as exc:
            self._raise_template_syntax_error(error=exc)
        except jinja2.TemplateNotFound as exc:
            raise JinjaTemplateNotFoundError(message=exc.message, filename=str(exc.name))

        return template

    def get_variables(self) -> list[str]:
        env = self.get_environment()

        template_source = self._template
        if self.is_file_based and env.loader:
            template_source = env.loader.get_source(env, self._template)[0]

        template = env.parse(template_source)

        return sorted(meta.find_undeclared_variables(template))

    def validate(self, restricted: bool = True) -> None:
        allowed_list = self._available_filters
        if restricted:
            allowed_list = self._trusted_filters

        env = self.get_environment()
        template_source = self._template
        if self.is_file_based and env.loader:
            template_source = env.loader.get_source(env, self._template)[0]

        template = env.parse(template_source)
        for node in template.find_all(nodes.Filter):
            if node.name not in allowed_list:
                raise JinjaTemplateOperationViolationError(f"The '{node.name}' filter isn't allowed to be used")

        forbidden_operations = ["Call", "Import", "Include"]
        if self.is_string_based and any(node.__class__.__name__ in forbidden_operations for node in template.body):
            raise JinjaTemplateOperationViolationError(
                f"These operations are forbidden for string based templates: {forbidden_operations}"
            )

    async def render(self, variables: dict[str, Any]) -> str:
        template = self.get_template()
        try:
            output = await template.render_async(variables)
        except jinja2.exceptions.TemplateNotFound as exc:
            raise JinjaTemplateNotFoundError(message=exc.message, filename=str(exc.name), base_template=template.name)
        except jinja2.TemplateSyntaxError as exc:
            self._raise_template_syntax_error(error=exc)
        except jinja2.UndefinedError as exc:
            traceback = Traceback(show_locals=False)
            errors = _identify_faulty_jinja_code(traceback=traceback)
            raise JinjaTemplateUndefinedError(message=exc.message, errors=errors)
        except Exception as exc:
            if error_message := getattr(exc, "message", None):
                message = error_message
            else:
                message = str(exc)
            raise JinjaTemplateError(message=message or "Unknown template error")

        return output

    def _get_string_based_environment(self) -> jinja2.Environment:
        env = SandboxedEnvironment(enable_async=True, undefined=jinja2.StrictUndefined)
        self._set_filters(env=env)
        self._environment = env
        return self._environment

    def _get_file_based_environment(self) -> jinja2.Environment:
        template_loader = jinja2.FileSystemLoader(searchpath=str(self._template_directory))
        env = jinja2.Environment(
            loader=template_loader,
            trim_blocks=True,
            lstrip_blocks=True,
            enable_async=True,
        )
        self._set_filters(env=env)
        self._environment = env
        return self._environment

    def _set_filters(self, env: jinja2.Environment) -> None:
        for default_filter in list(env.filters.keys()):
            if default_filter not in self._available_filters:
                del env.filters[default_filter]

        # Add filters from netutils
        env.filters.update(
            {name: jinja_filter for name, jinja_filter in netutils_filters.items() if name in self._available_filters}
        )
        # Add user supplied filters
        env.filters.update(self._filters)

    def _get_string_based_template(self) -> jinja2.Template:
        env = self.get_environment()
        self._template_definition = env.from_string(self._template)
        return self._template_definition

    def _get_file_based_template(self) -> jinja2.Template:
        env = self.get_environment()
        self._template_definition = env.get_template(self._template)
        return self._template_definition

    def _raise_template_syntax_error(self, error: jinja2.TemplateSyntaxError) -> NoReturn:
        filename: str | None = None
        if error.filename and self._template_directory:
            filename = error.filename
            if error.filename.startswith(str(self._template_directory)):
                filename = error.filename[len(str(self._template_directory)) :]

        raise JinjaTemplateSyntaxError(message=error.message, filename=filename, lineno=error.lineno)


def _identify_faulty_jinja_code(traceback: Traceback, nbr_context_lines: int = 3) -> list[UndefinedJinja2Error]:
    """This function identifies the faulty Jinja2 code and beautify it to provide meaningful information to the user.

    We use the rich's Traceback to parse the complete stack trace and extract Frames for each exception found in the trace.
    """
    response = []

    # Extract only the Jinja related exception
    for frame in [frame for frame in traceback.trace.stacks[0].frames if not frame.filename.endswith(".py")]:
        code = "".join(linecache.getlines(frame.filename))
        if frame.filename == "<template>":
            lexer_name = "text"
        else:
            lexer_name = Traceback._guess_lexer(frame.filename, code)
        syntax = Syntax(
            code,
            lexer_name,
            line_numbers=True,
            line_range=(
                frame.lineno - nbr_context_lines,
                frame.lineno + nbr_context_lines,
            ),
            highlight_lines={frame.lineno},
            code_width=88,
            theme=traceback.theme,
            dedent=False,
        )
        response.append(UndefinedJinja2Error(frame=frame, syntax=syntax))

    return response



from infrahub.core.validators.node.attribute import NodeAttributeAddChecker
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class SchemaNodeAttributeAddConstraintDependency(DependencyBuilder[NodeAttributeAddChecker]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> NodeAttributeAddChecker:
        return NodeAttributeAddChecker(db=context.db, branch=context.branch)



from __future__ import annotations

import hashlib
import importlib
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any

import ujson
import yaml
from infrahub_sdk import InfrahubClient  # noqa: TC002
from infrahub_sdk.exceptions import ValidationError
from infrahub_sdk.node import InfrahubNode
from infrahub_sdk.protocols import (
    CoreArtifact,
    CoreArtifactDefinition,
    CoreCheckDefinition,
    CoreGeneratorDefinition,
    CoreGenericRepository,
    CoreGraphQLQuery,
    CoreTransformation,
    CoreTransformJinja2,
    CoreTransformPython,
)
from infrahub_sdk.schema.repository import (
    InfrahubCheckDefinitionConfig,
    InfrahubGeneratorDefinitionConfig,
    InfrahubJinja2TransformConfig,
    InfrahubPythonTransformConfig,
    InfrahubRepositoryConfig,
)
from infrahub_sdk.spec.menu import MenuFile
from infrahub_sdk.spec.object import ObjectFile
from infrahub_sdk.template import Jinja2Template
from infrahub_sdk.template.exceptions import JinjaTemplateError
from infrahub_sdk.utils import compare_lists
from infrahub_sdk.yaml import InfrahubFile, SchemaFile
from prefect import flow, task
from prefect.cache_policies import NONE
from prefect.logging import get_run_logger
from pydantic import BaseModel, Field
from pydantic import ValidationError as PydanticValidationError
from typing_extensions import Self

from infrahub.core.constants import ArtifactStatus, ContentType, InfrahubKind, RepositoryObjects, RepositorySyncStatus
from infrahub.core.registry import registry
from infrahub.events.artifact_action import ArtifactCreatedEvent, ArtifactUpdatedEvent
from infrahub.events.models import EventMeta
from infrahub.events.repository_action import CommitUpdatedEvent
from infrahub.exceptions import CheckError, RepositoryInvalidFileSystemError, TransformError
from infrahub.git.base import InfrahubRepositoryBase, extract_repo_file_information
from infrahub.log import get_logger
from infrahub.workers.dependencies import get_event_service
from infrahub.workflows.utils import add_tags

if TYPE_CHECKING:
    import types

    from infrahub_sdk.checks import InfrahubCheck
    from infrahub_sdk.ctl.utils import YamlFileVar
    from infrahub_sdk.schema.repository import InfrahubRepositoryArtifactDefinitionConfig
    from infrahub_sdk.transforms import InfrahubTransform

    from infrahub.artifacts.models import CheckArtifactCreate
    from infrahub.git.models import RequestArtifactGenerate


class ArtifactGenerateResult(BaseModel):
    changed: bool
    checksum: str
    storage_id: str
    artifact_id: str


class InfrahubRepositoryJinja2(InfrahubJinja2TransformConfig):
    repository: str


class CheckDefinitionInformation(BaseModel):
    name: str
    """Name of the check"""

    repository: str = "self"
    """ID of the associated repository or self"""

    query: str
    """ID or name of the GraphQL Query associated with this Check"""

    file_path: str
    """Path to the python file within the repo"""

    class_name: str
    """Name of the Python Class"""

    check_class: Any
    """Python Class of the Check"""

    timeout: int
    """Timeout for the Check."""

    parameters: dict | None = None
    """Additional Parameters to extract from each target (if targets is provided)"""

    targets: str | None = Field(default=None, description="Targets if not a global check")


class TransformPythonInformation(BaseModel):
    name: str
    """Name of the Transform"""

    repository: str
    """ID or name of the repository this Transform is assocated with"""

    file_path: str
    """file_path of the TransformFunction within the repository"""

    query: str
    """ID or name of the GraphQLQuery this Transform is assocated with"""

    class_name: str
    """Name of the Python Class of the Transform Function"""

    transform_class: Any
    """Python Class of the Transform"""

    timeout: int
    """Timeout for the function."""

    convert_query_response: bool = Field(
        ..., description="Indicate if the transform should convert the query response to InfrahubNode objects"
    )


class InfrahubRepositoryIntegrator(InfrahubRepositoryBase):
    """
    This class provides interfaces to read and process information from .infrahub.yml files and can perform
    actions for objects defined within those files.

    This class will later be broken out from the "InfrahubRepository" based classes and instead be a separate
    class that uses an "InfrahubRepository" or "InfrahubReadOnlyRepository" as input
    """

    @classmethod
    async def init(cls, commit: str | None = None, **kwargs: Any) -> Self:
        self = cls(**kwargs)
        log = get_logger()
        try:
            self.validate_local_directories()
        except RepositoryInvalidFileSystemError:
            await self.ensure_location_is_defined()
            await self.create_locally(infrahub_branch_name=self.infrahub_branch_name, update_commit_value=False)
            log.info(f"Initialized the local directory for {self.name} because it was missing.")

        if commit:
            self.get_commit_worktree(commit=commit)

        log.debug(
            f"Initiated the object on an existing directory for {self.name}",
        )
        return self

    async def ensure_location_is_defined(self) -> None:
        if self.location:
            return
        client = self.sdk
        repo = await client.get(
            kind=CoreGenericRepository, name__value=self.name, exclude=["tags", "credential"], raise_when_missing=True
        )
        self.location = repo.location.value

    @flow(name="import-object-from-file", flow_run_name="Import objects")
    async def import_objects_from_files(
        self, infrahub_branch_name: str, git_branch_name: str | None = None, commit: str | None = None
    ) -> None:
        if not commit:
            commit = self.get_commit_value(branch_name=git_branch_name or infrahub_branch_name)

        await add_tags(branches=[infrahub_branch_name], nodes=[str(self.id)])

        self.create_commit_worktree(commit)
        await self._update_sync_status(branch_name=infrahub_branch_name, status=RepositorySyncStatus.SYNCING)

        config_file = await self.get_repository_config(branch_name=infrahub_branch_name, commit=commit)  # type: ignore[misc]
        sync_status = RepositorySyncStatus.IN_SYNC if config_file else RepositorySyncStatus.ERROR_IMPORT

        error: Exception | None = None

        try:
            if config_file:
                await self.import_schema_files(branch_name=infrahub_branch_name, commit=commit, config_file=config_file)  # type: ignore[misc]
                await self.import_all_graphql_query(
                    branch_name=infrahub_branch_name, commit=commit, config_file=config_file
                )  # type: ignore[misc]
                await self.import_objects(
                    branch_name=infrahub_branch_name,
                    commit=commit,
                    config_file=config_file,
                )  # type: ignore[misc]
                await self.import_all_python_files(  # type: ignore[call-overload]
                    branch_name=infrahub_branch_name, commit=commit, config_file=config_file
                )  # type: ignore[misc]
                await self.import_jinja2_transforms(
                    branch_name=infrahub_branch_name, commit=commit, config_file=config_file
                )  # type: ignore[misc]
                await self.import_artifact_definitions(
                    branch_name=infrahub_branch_name, commit=commit, config_file=config_file
                )  # type: ignore[misc]

        except Exception as exc:
            sync_status = RepositorySyncStatus.ERROR_IMPORT
            error = exc

        await self._update_sync_status(branch_name=infrahub_branch_name, status=sync_status)

        if error:
            raise error

        infrahub_branch = registry.get_branch_from_registry(branch=infrahub_branch_name)
        event_service = await get_event_service()
        await event_service.send(
            CommitUpdatedEvent(
                commit=commit,
                repository_name=self.name,
                repository_id=str(self.id),
                meta=EventMeta.with_dummy_context(branch=infrahub_branch),
            )
        )

    @task(name="import-jinja2-transforms", task_run_name="Import Jinja2 transform", cache_policy=NONE)
    async def import_jinja2_transforms(
        self,
        branch_name: str,
        commit: str,  # noqa: ARG002
        config_file: InfrahubRepositoryConfig,
    ) -> None:
        log = get_run_logger()

        schema = await self.sdk.schema.get(kind=InfrahubKind.TRANSFORMJINJA2, branch=branch_name)

        transforms_in_graph = {
            transform.name.value: transform
            for transform in await self.sdk.filters(
                kind=CoreTransformJinja2, branch=branch_name, repository__ids=[str(self.id)]
            )
        }

        local_transforms: dict[str, InfrahubRepositoryJinja2] = {}

        # Process the list of local Jinja2 Transforms to organize them by name
        log.info(f"Found {len(config_file.jinja2_transforms)} Jinja2 transforms in the repository")

        for config_transform in config_file.jinja2_transforms:
            try:
                self.sdk.schema.validate_data_against_schema(
                    schema=schema, data=config_transform.model_dump(exclude_none=True)
                )
            except PydanticValidationError as exc:
                for error in exc.errors():
                    locations = [str(error_location) for error_location in error["loc"]]
                    log.error(f"  {'/'.join(locations)} | {error['msg']} ({error['type']})")
                continue
            except ValidationError as exc:
                log.error(exc.message)
                continue

            transform = InfrahubRepositoryJinja2(repository=str(self.id), **config_transform.model_dump())

            # Query the GraphQL query and (eventually) replace the name with the ID
            graphql_query = await self.sdk.get(
                kind=InfrahubKind.GRAPHQLQUERY, branch=branch_name, id=str(transform.query), populate_store=True
            )
            transform.query = graphql_query.id

            local_transforms[transform.name] = transform

        present_in_both, only_graph, only_local = compare_lists(
            list1=list(transforms_in_graph.keys()), list2=list(local_transforms.keys())
        )

        for transform_name in only_local:
            log.info(f"New Jinja2 Transform {transform_name!r} found, creating")
            await self.create_jinja2_transform(branch_name=branch_name, data=local_transforms[transform_name])

        for transform_name in present_in_both:
            if not await self.compare_jinja2_transform(
                existing_transform=transforms_in_graph[transform_name], local_transform=local_transforms[transform_name]
            ):
                log.info(f"New version of the Jinja2 Transform '{transform_name}' found, updating")
                await self.update_jinja2_transform(
                    existing_transform=transforms_in_graph[transform_name],
                    local_transform=local_transforms[transform_name],
                )

        for transform_name in only_graph:
            log.info(f"Jinja2 Transform '{transform_name}' not found locally in branch {branch_name}, deleting")
            await transforms_in_graph[transform_name].delete()

    async def create_jinja2_transform(self, branch_name: str, data: InfrahubRepositoryJinja2) -> CoreTransformJinja2:
        schema = await self.sdk.schema.get(kind=InfrahubKind.TRANSFORMJINJA2, branch=branch_name)
        create_payload = self.sdk.schema.generate_payload_create(
            schema=schema, data=data.payload, source=self.id, is_protected=True
        )
        obj = await self.sdk.create(kind=CoreTransformJinja2, branch=branch_name, **create_payload)
        await obj.save()
        return obj

    @classmethod
    async def compare_jinja2_transform(
        cls, existing_transform: CoreTransformJinja2, local_transform: InfrahubRepositoryJinja2
    ) -> bool:
        if (
            existing_transform.description.value != local_transform.description
            or existing_transform.template_path.value != local_transform.template_path
            or existing_transform.query.id != local_transform.query
        ):
            return False

        return True

    async def update_jinja2_transform(
        self, existing_transform: CoreTransformJinja2, local_transform: InfrahubRepositoryJinja2
    ) -> None:
        if existing_transform.description.value != local_transform.description:
            existing_transform.description.value = local_transform.description

        if existing_transform.query.id != local_transform.query:
            existing_transform.query = {"id": local_transform.query, "source": str(self.id), "is_protected": True}

        if existing_transform.template_path.value != local_transform.template_path_value:
            existing_transform.template_path.value = local_transform.template_path_value

        await existing_transform.save()

    @task(name="import-artifact-definitions", task_run_name="Import Artifact Definitions", cache_policy=NONE)
    async def import_artifact_definitions(
        self,
        branch_name: str,
        commit: str,  # noqa: ARG002
        config_file: InfrahubRepositoryConfig,
    ) -> None:
        log = get_run_logger()
        schema = await self.sdk.schema.get(kind=InfrahubKind.ARTIFACTDEFINITION, branch=branch_name)

        artifact_defs_in_graph = {
            artdef.name.value: artdef
            for artdef in await self.sdk.filters(
                kind=CoreArtifactDefinition, branch=branch_name, prefetch_relationships=True, populate_store=True
            )
        }

        local_artifact_defs: dict[str, InfrahubRepositoryArtifactDefinitionConfig] = {}

        # Process the list of local Artifact Definitions to organize them by name
        log.info(f"Found {len(config_file.artifact_definitions)} artifact definitions in the repository")

        for artdef in config_file.artifact_definitions:
            try:
                self.sdk.schema.validate_data_against_schema(schema=schema, data=artdef.model_dump(exclude_none=True))
            except PydanticValidationError as exc:
                for error in exc.errors():
                    locations = [str(error_location) for error_location in error["loc"]]
                    log.error(f"  {'/'.join(locations)} | {error['msg']} ({error['type']})")
                continue
            except ValidationError as exc:
                log.error(exc.message)
                continue

            local_artifact_defs[artdef.name] = artdef

        present_in_both, _, only_local = compare_lists(
            list1=list(artifact_defs_in_graph.keys()), list2=list(local_artifact_defs.keys())
        )

        for artdef_name in only_local:
            log.info(f"New Artifact Definition {artdef_name!r} found, creating")
            await self.create_artifact_definition(branch_name=branch_name, data=local_artifact_defs[artdef_name])

        for artdef_name in present_in_both:
            if not await self.compare_artifact_definition(
                existing_artifact_definition=artifact_defs_in_graph[artdef_name],
                local_artifact_definition=local_artifact_defs[artdef_name],
            ):
                log.info(f"New version of the Artifact Definition '{artdef_name}' found, updating")
                await self.update_artifact_definition(
                    existing_artifact_definition=artifact_defs_in_graph[artdef_name],
                    local_artifact_definition=local_artifact_defs[artdef_name],
                )

    async def create_artifact_definition(
        self, branch_name: str, data: InfrahubRepositoryArtifactDefinitionConfig
    ) -> InfrahubNode:
        schema = await self.sdk.schema.get(kind=InfrahubKind.ARTIFACTDEFINITION, branch=branch_name)
        create_payload = self.sdk.schema.generate_payload_create(
            schema=schema, data=data.model_dump(), source=self.id, is_protected=True
        )
        obj = await self.sdk.create(kind=InfrahubKind.ARTIFACTDEFINITION, branch=branch_name, **create_payload)
        await obj.save()
        return obj

    @classmethod
    async def compare_artifact_definition(
        cls,
        existing_artifact_definition: CoreArtifactDefinition,
        local_artifact_definition: InfrahubRepositoryArtifactDefinitionConfig,
    ) -> bool:
        if (
            existing_artifact_definition.artifact_name.value != local_artifact_definition.artifact_name
            or existing_artifact_definition.parameters.value != local_artifact_definition.parameters
            or existing_artifact_definition.content_type.value != local_artifact_definition.content_type
            or existing_artifact_definition.targets.peer.name.value != local_artifact_definition.targets
        ):
            return False

        return True

    async def update_artifact_definition(
        self,
        existing_artifact_definition: CoreArtifactDefinition,
        local_artifact_definition: InfrahubRepositoryArtifactDefinitionConfig,
    ) -> None:
        if existing_artifact_definition.artifact_name.value != local_artifact_definition.artifact_name:
            existing_artifact_definition.artifact_name.value = local_artifact_definition.artifact_name

        if existing_artifact_definition.parameters.value != local_artifact_definition.parameters:
            existing_artifact_definition.parameters.value = local_artifact_definition.parameters

        if existing_artifact_definition.content_type.value != local_artifact_definition.content_type:
            existing_artifact_definition.content_type.value = local_artifact_definition.content_type

        if existing_artifact_definition.targets.peer.name.value != local_artifact_definition.targets:
            existing_artifact_definition.targets = local_artifact_definition.targets

        await existing_artifact_definition.save()

    @task(name="repository-get-config", task_run_name="get repository config", cache_policy=NONE)
    async def get_repository_config(self, branch_name: str, commit: str) -> InfrahubRepositoryConfig | None:
        branch_wt = self.get_worktree(identifier=commit or branch_name)
        log = get_run_logger()

        # Check for both .infrahub.yml and .infrahub.yaml, prefer .yml if both exist
        config_file_yml = branch_wt.directory / ".infrahub.yml"
        config_file_yaml = branch_wt.directory / ".infrahub.yaml"

        if config_file_yml.is_file():
            config_file = config_file_yml
            config_file_name = ".infrahub.yml"
        elif config_file_yaml.is_file():
            config_file = config_file_yaml
            config_file_name = ".infrahub.yaml"
        else:
            log.debug("Unable to find the configuration file (.infrahub.yml or .infrahub.yaml), skipping")
            return None

        config_file_content = config_file.read_text(encoding="utf-8")
        try:
            data = yaml.safe_load(config_file_content)
        except yaml.YAMLError as exc:
            log.error(f"Unable to load the configuration file in YAML format {config_file_name} : {exc}")
            return None

        # Convert data to a dictionary to avoid it being `None` if the yaml file is just an empty document
        data = data or {}

        try:
            configuration = InfrahubRepositoryConfig(**data)
            log.info(f"Successfully parsed {config_file_name}")
            return configuration
        except PydanticValidationError as exc:
            log.error(f"Unable to load the configuration file {config_file_name}, the format is not valid  : {exc}")
            return None

    @task(name="import-schema-files", task_run_name="Import schema files", cache_policy=NONE)
    async def import_schema_files(self, branch_name: str, commit: str, config_file: InfrahubRepositoryConfig) -> None:
        log = get_run_logger()
        branch_wt = self.get_worktree(identifier=commit or branch_name)

        schemas_data: list[SchemaFile] = []

        for schema in config_file.schemas:
            full_schema = branch_wt.directory / schema
            if not full_schema.exists():
                log.warning(f"Unable to find the schema {schema}")

            if full_schema.is_file():
                schema_file = SchemaFile(identifier=str(schema), location=full_schema)
                schema_file.load_content()
                schemas_data.append(schema_file)
            elif full_schema.is_dir():
                files = await self.find_files(
                    extension=["yaml", "yml", "json"], branch_name=branch_name, commit=commit, directory=full_schema
                )
                for item in files:
                    identifier = str(item.relative_to(branch_wt.directory))
                    schema_file = SchemaFile(identifier=identifier, location=item)
                    schema_file.load_content()
                    schemas_data.append(schema_file)

        if not schemas_data:
            # If the repository doesn't contain any schema files there is no reason to continue
            # and send an empty list to the API
            return

        for schema_file in schemas_data:
            if schema_file.valid:
                continue
            log.error(f"Unable to load the file {schema_file.identifier}, {schema_file.error_message}")

        # Valid data format of content
        for schema_file in schemas_data:
            try:
                self.sdk.schema.validate(schema_file.content)
            except PydanticValidationError as exc:
                log.error(f"Schema not valid, found '{len(exc.errors())}' error(s) in {schema_file.identifier} : {exc}")
                raise ValidationError(
                    identifier=str(self.id),
                    message=f"Schema not valid, found '{len(exc.errors())}' error(s) in {schema_file.identifier} : {exc}",
                ) from exc

        response = await self.sdk.schema.load(
            schemas=[item.content for item in schemas_data], branch=branch_name, wait_until_converged=True
        )

        if response.errors:
            error_messages = []

            if "detail" in response.errors:
                for error in response.errors["detail"]:
                    loc_str = [str(item) for item in error["loc"][1:]]
                    error_messages.append(f"{'/'.join(loc_str)} | {error['msg']} ({error['type']})")
            elif "error" in response.errors:
                error_messages.append(f"{response.errors.get('error')}")
            else:
                error_messages.append(f"{response.errors}")

            log.error(f"Unable to load the schema : {', '.join(error_messages)}")

            raise ValidationError(
                identifier=str(self.id), message=f"Unable to load the schema : {', '.join(error_messages)}"
            )

        for schema_file in schemas_data:
            log.info(f"schema '{schema_file.identifier}' loaded successfully!")

    @task(name="import-graphql-queries", task_run_name="Import GraphQL Queries", cache_policy=NONE)
    async def import_all_graphql_query(
        self, branch_name: str, commit: str, config_file: InfrahubRepositoryConfig
    ) -> None:
        """Search for all .gql file and import them as GraphQL query."""
        log = get_run_logger()

        commit_wt = self.get_worktree(identifier=commit)
        local_queries = {
            query.name: query.load_query(relative_path=commit_wt.directory) for query in config_file.queries
        }

        if not local_queries:
            return

        queries_in_graph = {
            query.name.value: query
            for query in await self.sdk.filters(
                kind=CoreGraphQLQuery, branch=branch_name, repository__ids=[str(self.id)]
            )
        }

        present_in_both, only_graph, only_local = compare_lists(
            list1=list(queries_in_graph.keys()), list2=list(local_queries.keys())
        )

        for query_name in only_local:
            query = local_queries[query_name]
            log.info(f"New Graphql Query {query_name!r} found, creating")
            await self.create_graphql_query(branch_name=branch_name, name=query_name, query_string=query)

        for query_name in present_in_both:
            local_query = local_queries[query_name]
            graph_query = queries_in_graph[query_name]
            if local_query != graph_query.query.value:
                log.info(f"New version of the Graphql Query {query_name!r} found, updating")
                graph_query.query.value = local_query
                await graph_query.save()

        for query_name in only_graph:
            graph_query = queries_in_graph[query_name]
            log.info(f"Graphql Query {query_name!r} not found locally, deleting")
            await graph_query.delete()

    async def create_graphql_query(self, branch_name: str, name: str, query_string: str) -> CoreGraphQLQuery:
        data = {"name": name, "query": query_string, "repository": self.id}

        schema = await self.sdk.schema.get(kind=InfrahubKind.GRAPHQLQUERY, branch=branch_name)
        create_payload = self.sdk.schema.generate_payload_create(
            schema=schema,
            data=data,
            source=self.id,
            is_protected=True,
        )
        obj = await self.sdk.create(kind=CoreGraphQLQuery, branch=branch_name, **create_payload)
        await obj.save()
        return obj

    @task(name="import-python-check-definitions", task_run_name="Import Python Check Definitions", cache_policy=NONE)
    async def import_python_check_definitions(
        self, branch_name: str, commit: str, config_file: InfrahubRepositoryConfig
    ) -> None:
        log = get_run_logger()

        commit_wt = self.get_worktree(identifier=commit)
        branch_wt = self.get_worktree(identifier=commit or branch_name)

        # Ensure the path for this repository is present in sys.path
        if str(self.directory_root) not in sys.path:
            sys.path.append(str(self.directory_root))

        checks = []
        log.info(f"Found {len(config_file.check_definitions)} check definitions in the repository")
        for check in config_file.check_definitions:
            log.debug(f"{self.name}, file={check.file_path}")

            file_info = extract_repo_file_information(
                full_filename=branch_wt.directory / check.file_path,
                repo_directory=self.directory_root,
                worktree_directory=commit_wt.directory,
            )
            try:
                module = importlib.import_module(file_info.module_name)
            except ModuleNotFoundError as exc:
                log.warning(f"{self.name},  file={check.file_path.as_posix()} error={str(exc)}")
                raise

            checks.extend(
                await self.get_check_definition(
                    branch_name=branch_name,
                    module=module,
                    file_path=file_info.relative_path_file,
                    check_definition=check,
                )  # type: ignore[misc]
            )

        local_check_definitions = {check.name: check for check in checks}
        check_definition_in_graph = {
            check.name.value: check
            for check in await self.sdk.filters(
                kind=CoreCheckDefinition, branch=branch_name, repository__ids=[str(self.id)]
            )
        }

        present_in_both, only_graph, only_local = compare_lists(
            list1=list(check_definition_in_graph.keys()), list2=list(local_check_definitions.keys())
        )

        for check_name in only_local:
            log.info(f"New CheckDefinition {check_name!r} found, creating")
            await self.create_python_check_definition(
                branch_name=branch_name, check=local_check_definitions[check_name]
            )

        for check_name in present_in_both:
            if not await self.compare_python_check_definition(
                check=local_check_definitions[check_name],
                existing_check=check_definition_in_graph[check_name],
            ):
                log.info(f"New version of CheckDefinition {check_name!r} found, updating")
                await self.update_python_check_definition(
                    check=local_check_definitions[check_name],
                    existing_check=check_definition_in_graph[check_name],
                )

        for check_name in only_graph:
            log.info(f"CheckDefinition '{check_name!r}' not found locally, deleting")
            await check_definition_in_graph[check_name].delete()

    @task(name="import-generator-definitions", task_run_name="Import Generator Definitions", cache_policy=NONE)
    async def import_generator_definitions(
        self, branch_name: str, commit: str, config_file: InfrahubRepositoryConfig
    ) -> None:
        log = get_run_logger()

        commit_wt = self.get_worktree(identifier=commit)
        branch_wt = self.get_worktree(identifier=commit or branch_name)

        generators = []
        log.info(f"Found {len(config_file.generator_definitions)} generator definitions in the repository")

        for generator in config_file.generator_definitions:
            log.info(f"Processing generator {generator.name} ({generator.file_path})")
            file_info = extract_repo_file_information(
                full_filename=branch_wt.directory / generator.file_path,
                repo_directory=self.directory_root,
                worktree_directory=commit_wt.directory,
            )

            generator.load_class(import_root=self.directory_root, relative_path=file_info.relative_repo_path_dir)
            generators.append(generator)

        local_generator_definitions = {generator.name: generator for generator in generators}
        generator_definition_in_graph = {
            generator.name.value: generator
            for generator in await self.sdk.filters(
                kind=CoreGeneratorDefinition, branch=branch_name, repository__ids=[str(self.id)]
            )
        }

        present_in_both, only_graph, only_local = compare_lists(
            list1=list(generator_definition_in_graph.keys()), list2=list(local_generator_definitions.keys())
        )

        for generator_name in only_local:
            log.info(f"New GeneratorDefinition {generator_name!r} found, creating")
            await self._create_generator_definition(
                branch_name=branch_name, generator=local_generator_definitions[generator_name]
            )

        for generator_name in present_in_both:
            if await self._generator_requires_update(
                generator=local_generator_definitions[generator_name],
                existing_generator=generator_definition_in_graph[generator_name],
                branch_name=branch_name,
            ):
                log.info(f"New version of GeneratorDefinition {generator_name!r} found, updating")

                await self._update_generator_definition(
                    generator=local_generator_definitions[generator_name],
                    existing_generator=generator_definition_in_graph[generator_name],
                )

        for generator_name in only_graph:
            log.info(f"GeneratorDefinition '{generator_name!r}' not found locally, deleting")
            await generator_definition_in_graph[generator_name].delete()

    async def _generator_requires_update(
        self,
        generator: InfrahubGeneratorDefinitionConfig,
        existing_generator: CoreGeneratorDefinition,
        branch_name: str,
    ) -> bool:
        graphql_queries = await self.sdk.filters(
            kind=InfrahubKind.GRAPHQLQUERY, branch=branch_name, name__value=generator.query, populate_store=True
        )
        if graphql_queries:
            generator.query = graphql_queries[0].id
        targets = await self.sdk.filters(
            kind=InfrahubKind.GENERICGROUP,
            branch=branch_name,
            name__value=generator.targets,
            populate_store=True,
            fragment=True,
        )
        if targets:
            generator.targets = targets[0].id

        if (
            existing_generator.query.id != generator.query
            or existing_generator.file_path.value != str(generator.file_path)
            or existing_generator.class_name.value != generator.class_name
            or existing_generator.parameters.value != generator.parameters
            or existing_generator.convert_query_response.value != generator.convert_query_response
            or existing_generator.targets.id != generator.targets
            or existing_generator.execute_in_proposed_change.value != generator.execute_in_proposed_change
            or existing_generator.execute_after_merge.value != generator.execute_after_merge
        ):
            return True
        return False

    @task(name="import-python-transforms", task_run_name="Import Python Transforms", cache_policy=NONE)
    async def import_python_transforms(
        self, branch_name: str, commit: str, config_file: InfrahubRepositoryConfig
    ) -> None:
        log = get_run_logger()
        commit_wt = self.get_worktree(identifier=commit)
        branch_wt = self.get_worktree(identifier=commit or branch_name)

        # Ensure the path for this repository is present in sys.path
        if str(self.directory_root) not in sys.path:
            sys.path.append(str(self.directory_root))

        transforms: list[TransformPythonInformation] = []
        log.info(f"Found {len(config_file.python_transforms)} Python transforms in the repository")

        for transform in config_file.python_transforms:
            log.debug(f"{self.name}, file={transform.file_path}")

            file_info = extract_repo_file_information(
                full_filename=branch_wt.directory / transform.file_path,
                repo_directory=self.directory_root,
                worktree_directory=commit_wt.directory,
            )
            try:
                module = importlib.import_module(file_info.module_name)
            except ModuleNotFoundError as exc:
                log.warning(f"{self.name}, file={transform.file_path.as_posix()} error={str(exc)}")
                raise

            transforms.extend(
                await self.get_python_transforms(
                    branch_name=branch_name,
                    module=module,
                    file_path=file_info.relative_path_file,
                    transform=transform,
                )  # type: ignore[misc]
            )

        local_transform_definitions = {transform.name: transform for transform in transforms}
        transform_definition_in_graph = {
            transform.name.value: transform
            for transform in await self.sdk.filters(
                kind=CoreTransformPython, branch=branch_name, repository__ids=[str(self.id)]
            )
        }

        present_in_both, only_graph, only_local = compare_lists(
            list1=list(transform_definition_in_graph.keys()), list2=list(local_transform_definitions.keys())
        )

        for transform_name in only_local:
            log.info(f"New TransformPython {transform_name!r} found, creating")
            await self.create_python_transform(
                branch_name=branch_name, transform=local_transform_definitions[transform_name]
            )

        for transform_name in present_in_both:
            if not await self.compare_python_transform(
                local_transform=local_transform_definitions[transform_name],
                existing_transform=transform_definition_in_graph[transform_name],
            ):
                log.info(f"New version of TransformPython {transform_name!r} found, updating")
                await self.update_python_transform(
                    local_transform=local_transform_definitions[transform_name],
                    existing_transform=transform_definition_in_graph[transform_name],
                )

        for transform_name in only_graph:
            log.info(f"TransformPython {transform_name!r} not found locally, deleting")
            await transform_definition_in_graph[transform_name].delete()

    async def _load_yamlfile_from_disk(self, paths: list[Path], file_type: type[YamlFileVar]) -> list[YamlFileVar]:
        data_files = file_type.load_from_disk(paths=paths)

        for data_file in data_files:
            if not data_file.valid or not data_file.content:
                raise ValueError(f"{data_file.error_message} ({data_file.location})")

        return data_files

    async def _load_objects(
        self,
        paths: list[Path],
        branch: str,
        file_type: type[InfrahubFile],
    ) -> None:
        """Load one or multiple objects files into Infrahub."""

        log = get_run_logger()
        files = await self._load_yamlfile_from_disk(paths=paths, file_type=file_type)

        for file in files:
            await file.validate_format(client=self.sdk, branch=branch)
            schema = await self.sdk.schema.get(kind=file.spec.kind, branch=branch)
            if not schema.human_friendly_id and not schema.default_filter:
                raise ValueError(
                    f"Schemas of objects or menus defined within {file.location} "
                    "should have a `human_friendly_id` defined to avoid creating duplicated objects."
                )

        for file in files:
            log.info(f"Loading objects defined in {file.location}")
            await file.process(client=self.sdk, branch=branch)

    async def _import_file_paths(
        self, branch_name: str, commit: str, files_pathes: list[Path], object_type: RepositoryObjects
    ) -> None:
        branch_wt = self.get_worktree(identifier=commit or branch_name)
        file_pathes = [branch_wt.directory / file_path for file_path in files_pathes]

        # We currently assume there can't be concurrent imports, but if so, we might need to clone the client before tracking here.
        async with self.sdk.start_tracking(
            identifier=f"group-repo-{object_type.value}-{self.id}",
            delete_unused_nodes=True,
            branch=branch_name,
            group_type="CoreRepositoryGroup",
            group_params={"content": object_type.value, "repository": str(self.id)},
        ):
            file_type = repo_object_type_to_file_type(object_type)
            await self._load_objects(
                paths=file_pathes,
                branch=branch_name,
                file_type=file_type,
            )

    @task(name="import-objects", task_run_name="Import Objects", cache_policy=NONE)
    async def import_objects(
        self,
        branch_name: str,
        commit: str,
        config_file: InfrahubRepositoryConfig,
    ) -> None:
        await self._import_file_paths(
            branch_name=branch_name,
            commit=commit,
            files_pathes=config_file.objects,
            object_type=RepositoryObjects.OBJECT,
        )
        await self._import_file_paths(
            branch_name=branch_name,
            commit=commit,
            files_pathes=config_file.menus,
            object_type=RepositoryObjects.MENU,
        )

    @task(name="check-definition-get", task_run_name="Get Check Definition", cache_policy=NONE)
    async def get_check_definition(
        self,
        branch_name: str,
        module: types.ModuleType,
        file_path: str,
        check_definition: InfrahubCheckDefinitionConfig,
    ) -> list[CheckDefinitionInformation]:
        log = get_run_logger()
        if check_definition.class_name not in dir(module):
            return []

        checks = []
        check_class = getattr(module, check_definition.class_name)

        try:
            graphql_query = await self.sdk.get(
                kind=InfrahubKind.GRAPHQLQUERY, branch=branch_name, id=str(check_class.query), populate_store=True
            )
            checks.append(
                CheckDefinitionInformation(
                    name=check_definition.name,
                    repository=str(self.id),
                    class_name=check_definition.class_name,
                    check_class=check_class,
                    file_path=file_path,
                    query=str(graphql_query.id),
                    timeout=check_class.timeout,
                    parameters=check_definition.parameters,
                    targets=check_definition.targets,
                )
            )

        except Exception as exc:
            log.error(
                f"An error occurred while processing the CheckDefinition {check_class.__name__} from {file_path} : {exc} "
            )
            raise
        return checks

    @task(name="python-transform-get", task_run_name="Get Python Transform", cache_policy=NONE)
    async def get_python_transforms(
        self, branch_name: str, module: types.ModuleType, file_path: str, transform: InfrahubPythonTransformConfig
    ) -> list[TransformPythonInformation]:
        log = get_run_logger()
        if transform.class_name not in dir(module):
            return []

        transforms = []
        transform_class = getattr(module, transform.class_name)
        graphql_query = await self.sdk.get(
            kind=InfrahubKind.GRAPHQLQUERY, branch=branch_name, id=str(transform_class.query), populate_store=True
        )
        try:
            transforms.append(
                TransformPythonInformation(
                    name=transform.name,
                    repository=str(self.id),
                    class_name=transform.class_name,
                    transform_class=transform_class,
                    file_path=file_path,
                    query=str(graphql_query.id),
                    timeout=transform_class.timeout,
                    convert_query_response=transform.convert_query_response,
                )
            )

        except Exception as exc:
            log.error(
                f"An error occurred while processing the PythonTransform {transform.name} from {file_path} : {exc} "
            )
            raise

        return transforms

    async def _create_generator_definition(
        self, generator: InfrahubGeneratorDefinitionConfig, branch_name: str
    ) -> InfrahubNode:
        data = generator.model_dump(exclude_none=True, exclude={"file_path"})
        data["file_path"] = str(generator.file_path)
        data["repository"] = self.id

        schema = await self.sdk.schema.get(kind=InfrahubKind.GENERATORDEFINITION, branch=branch_name)

        create_payload = self.sdk.schema.generate_payload_create(
            schema=schema,
            data=data,
            source=str(self.id),
            is_protected=True,
        )
        obj = await self.sdk.create(kind=InfrahubKind.GENERATORDEFINITION, branch=branch_name, **create_payload)
        await obj.save()

        return obj

    async def _update_generator_definition(
        self,
        generator: InfrahubGeneratorDefinitionConfig,
        existing_generator: CoreGeneratorDefinition,
    ) -> None:
        if existing_generator.query.id != generator.query:
            existing_generator.query = {"id": generator.query, "source": str(self.id), "is_protected": True}

        if existing_generator.class_name.value != generator.class_name:
            existing_generator.class_name.value = generator.class_name

        if existing_generator.file_path.value != str(generator.file_path):
            existing_generator.file_path.value = str(generator.file_path)

        if existing_generator.convert_query_response.value != generator.convert_query_response:
            existing_generator.convert_query_response.value = generator.convert_query_response

        if existing_generator.parameters.value != generator.parameters:
            existing_generator.parameters.value = generator.parameters

        if existing_generator.targets.id != generator.targets:
            existing_generator.targets = {"id": generator.targets, "source": str(self.id), "is_protected": True}

        if existing_generator.execute_in_proposed_change.value != generator.execute_in_proposed_change:
            existing_generator.execute_in_proposed_change.value = generator.execute_in_proposed_change

        if existing_generator.execute_after_merge.value != generator.execute_after_merge:
            existing_generator.execute_after_merge.value = generator.execute_after_merge

        await existing_generator.save()

    async def create_python_check_definition(
        self, branch_name: str, check: CheckDefinitionInformation
    ) -> CoreCheckDefinition:
        data = {
            "name": check.name,
            "repository": check.repository,
            "query": check.query,
            "file_path": check.file_path,
            "class_name": check.class_name,
            "timeout": check.timeout,
            "parameters": check.parameters,
        }

        if check.targets:
            data["targets"] = check.targets

        schema = await self.sdk.schema.get(kind=InfrahubKind.CHECKDEFINITION, branch=branch_name)

        create_payload = self.sdk.schema.generate_payload_create(
            schema=schema,
            data=data,
            source=self.id,
            is_protected=True,
        )
        obj = await self.sdk.create(kind=CoreCheckDefinition, branch=branch_name, data=create_payload)
        await obj.save()

        return obj

    async def update_python_check_definition(
        self,
        check: CheckDefinitionInformation,
        existing_check: CoreCheckDefinition,
    ) -> None:
        if existing_check.query.id != check.query:
            existing_check.query = {"id": check.query, "source": str(self.id), "is_protected": True}

        if existing_check.file_path.value != check.file_path:
            existing_check.file_path.value = check.file_path

        if existing_check.timeout.value != check.timeout:
            existing_check.timeout.value = check.timeout

        if existing_check.parameters.value != check.parameters:
            existing_check.parameters.value = check.parameters

        await existing_check.save()

    @classmethod
    async def compare_python_check_definition(
        cls, check: CheckDefinitionInformation, existing_check: CoreCheckDefinition
    ) -> bool:
        """Compare an existing Python Check Object with a Check Class
        and identify if we need to update the object in the database."""
        if (
            existing_check.query.id != check.query
            or existing_check.file_path.value != check.file_path
            or existing_check.timeout.value != check.timeout
            or existing_check.class_name.value != check.class_name
            or existing_check.parameters.value != check.parameters
        ):
            return False
        return True

    async def create_python_transform(
        self, branch_name: str, transform: TransformPythonInformation
    ) -> CoreTransformPython:
        schema = await self.sdk.schema.get(kind=InfrahubKind.TRANSFORMPYTHON, branch=branch_name)
        data = {
            "name": transform.name,
            "repository": transform.repository,
            "query": transform.query,
            "file_path": transform.file_path,
            "class_name": transform.class_name,
            "timeout": transform.timeout,
            "convert_query_response": transform.convert_query_response,
        }
        create_payload = self.sdk.schema.generate_payload_create(
            schema=schema,
            data=data,
            source=str(self.id),
            is_protected=True,
        )
        obj = await self.sdk.create(kind=CoreTransformPython, branch=branch_name, data=create_payload)
        await obj.save()
        return obj

    async def update_python_transform(
        self, existing_transform: CoreTransformPython, local_transform: TransformPythonInformation
    ) -> None:
        if existing_transform.query.id != local_transform.query:
            existing_transform.query = {"id": local_transform.query, "source": str(self.id), "is_protected": True}

        if existing_transform.file_path.value != local_transform.file_path:
            existing_transform.file_path.value = local_transform.file_path

        if existing_transform.timeout.value != local_transform.timeout:
            existing_transform.timeout.value = local_transform.timeout

        if existing_transform.convert_query_response.value != local_transform.convert_query_response:
            existing_transform.convert_query_response.value = local_transform.convert_query_response

        await existing_transform.save()

    @classmethod
    async def compare_python_transform(
        cls, existing_transform: CoreTransformPython, local_transform: TransformPythonInformation
    ) -> bool:
        if (
            existing_transform.query.id != local_transform.query
            or existing_transform.file_path.value != local_transform.file_path
            or existing_transform.timeout.value != local_transform.timeout
            or existing_transform.convert_query_response.value != local_transform.convert_query_response
        ):
            return False
        return True

    @flow(name="import-python-files", flow_run_name="Import Python file")
    async def import_all_python_files(
        self, branch_name: str, commit: str, config_file: InfrahubRepositoryConfig
    ) -> None:
        await add_tags(branches=[branch_name], nodes=[str(self.id)])

        await self.import_python_check_definitions(branch_name=branch_name, commit=commit, config_file=config_file)  # type: ignore[misc]
        await self.import_python_transforms(branch_name=branch_name, commit=commit, config_file=config_file)  # type: ignore[misc]
        await self.import_generator_definitions(branch_name=branch_name, commit=commit, config_file=config_file)  # type: ignore[misc]

    @task(name="jinja2-template-render", task_run_name="Render Jinja2 template", cache_policy=NONE)
    async def render_jinja2_template(self, commit: str, location: str, data: dict) -> str:
        log = get_run_logger()
        commit_worktree = self.get_commit_worktree(commit=commit)

        self.validate_location(commit=commit, worktree_directory=commit_worktree.directory, file_path=location)

        jinja2_template = Jinja2Template(template=Path(location), template_directory=Path(commit_worktree.directory))
        try:
            return await jinja2_template.render(variables=data)
        except JinjaTemplateError as exc:
            log.error(str(exc), exc_info=True)
            raise TransformError(
                repository_name=self.name, commit=commit, location=location, message=exc.message
            ) from exc

    @task(name="python-check-execute", task_run_name="Execute Python Check", cache_policy=NONE)
    async def execute_python_check(
        self,
        branch_name: str,
        commit: str,
        location: str,
        class_name: str,
        client: InfrahubClient,
        params: dict | None = None,
    ) -> InfrahubCheck:
        """Execute A Python Check stored in the repository."""
        log = get_run_logger()

        commit_worktree = self.get_commit_worktree(commit=commit)

        self.validate_location(commit=commit, worktree_directory=commit_worktree.directory, file_path=location)

        # Ensure the path for this repository is present in sys.path
        if str(self.directory_root) not in sys.path:
            sys.path.append(str(self.directory_root))

        try:
            file_info = extract_repo_file_information(
                full_filename=commit_worktree.directory / location,
                repo_directory=self.directory_root,
                worktree_directory=commit_worktree.directory,
            )

            module = importlib.import_module(file_info.module_name)

            check_class: type[InfrahubCheck] = getattr(module, class_name)

            check = check_class(
                root_directory=commit_worktree.directory, branch=branch_name, client=client, params=params
            )
            await check.run()

            return check

        except ModuleNotFoundError as exc:
            error_msg = "Unable to load the check file"
            log.error(error_msg)
            raise CheckError(
                repository_name=self.name, class_name=class_name, commit=commit, location=location, message=error_msg
            ) from exc

        except AttributeError as exc:
            error_msg = f"Unable to find the class {class_name}"
            log.error(error_msg)
            raise CheckError(
                repository_name=self.name, class_name=class_name, commit=commit, location=location, message=error_msg
            ) from exc

        except Exception as exc:
            log.critical(str(exc), exc_info=True)
            raise CheckError(
                repository_name=self.name, class_name=class_name, commit=commit, location=location, message=str(exc)
            ) from exc

    @task(name="python-transform-execute", task_run_name="Execute Python Transform", cache_policy=NONE)
    async def execute_python_transform(
        self,
        branch_name: str,
        commit: str,
        location: str,
        client: InfrahubClient,
        convert_query_response: bool,
        data: dict | None = None,
    ) -> Any:
        """Execute A Python Transform stored in the repository."""
        log = get_run_logger()

        if "::" not in location:
            raise ValueError("Transformation location not valid, it must contains a double colons (::)")

        file_path, class_name = location.split("::")
        commit_worktree = self.get_commit_worktree(commit=commit)

        log.debug(f"Will run Python Transform from {class_name} at {location}")

        self.validate_location(commit=commit, worktree_directory=commit_worktree.directory, file_path=file_path)

        # Ensure the path for this repository is present in sys.path
        if str(self.directory_root) not in sys.path:
            sys.path.append(str(self.directory_root))

        try:
            file_info = extract_repo_file_information(
                full_filename=commit_worktree.directory / file_path,
                repo_directory=self.directory_root,
                worktree_directory=commit_worktree.directory,
            )

            module = importlib.import_module(file_info.module_name)

            transform_class: type[InfrahubTransform] = getattr(module, class_name)

            transform = transform_class(
                root_directory=commit_worktree.directory,
                branch=branch_name,
                client=client,
                convert_query_response=convert_query_response,
                infrahub_node=InfrahubNode,
            )
            return await transform.run(data=data)
        except ModuleNotFoundError as exc:
            error_msg = f"Unable to load the transform file {location}"
            log.error(error_msg)
            raise TransformError(
                repository_name=self.name, commit=commit, location=location, message=error_msg
            ) from exc

        except AttributeError as exc:
            error_msg = f"Unable to find the class {class_name} in {location}"
            log.error(error_msg)
            raise TransformError(
                repository_name=self.name, commit=commit, location=location, message=error_msg
            ) from exc

        except Exception as exc:
            log.critical(str(exc), exc_info=True)
            raise TransformError(repository_name=self.name, commit=commit, location=location, message=str(exc)) from exc

    async def artifact_generate(
        self,
        branch_name: str,
        commit: str,
        artifact: CoreArtifact,
        target: InfrahubNode,
        definition: CoreArtifactDefinition,
        transformation: CoreTransformation,
        query: CoreGraphQLQuery,
    ) -> ArtifactGenerateResult:
        """It doesn't look like this is used anywhere today ... we should either remove it or refactor render_artifact below to use this."""
        variables = await target.extract(params=definition.parameters.value)
        response = await self.sdk.query_gql_query(
            name=query.name.value,
            variables=variables,
            update_group=True,
            subscribers=[artifact.id],
            tracker="artifact-query-graphql-data",
            branch_name=branch_name,
            timeout=transformation.timeout.value,
        )

        if transformation.typename == InfrahubKind.TRANSFORMJINJA2:
            artifact_content = await self.render_jinja2_template.with_options(
                timeout_seconds=transformation.timeout.value
            )(commit=commit, location=transformation.template_path.value, data=response)  # type: ignore[misc]
        elif transformation.typename == InfrahubKind.TRANSFORMPYTHON:
            transformation_location = f"{transformation.file_path.value}::{transformation.class_name.value}"
            artifact_content = await self.execute_python_transform.with_options(
                timeout_seconds=transformation.timeout.value
            )(
                client=self.sdk,
                branch_name=branch_name,
                commit=commit,
                location=transformation_location,
                data=response,
                convert_query_response=transformation.convert_query_response.value,
            )  # type: ignore[misc]

        if definition.content_type.value == ContentType.APPLICATION_JSON.value and isinstance(artifact_content, dict):
            artifact_content_str = ujson.dumps(artifact_content, indent=2)
        elif definition.content_type.value == ContentType.APPLICATION_YAML.value and isinstance(artifact_content, dict):
            artifact_content_str = yaml.dump(artifact_content, indent=2)
        else:
            artifact_content_str = str(artifact_content)

        checksum = hashlib.md5(bytes(artifact_content_str, encoding="utf-8"), usedforsecurity=False).hexdigest()

        if artifact.checksum.value == checksum:
            return ArtifactGenerateResult(
                changed=False, checksum=checksum, storage_id=artifact.storage_id.value, artifact_id=artifact.id
            )

        resp = await self.sdk.object_store.upload(content=artifact_content_str, tracker="artifact-upload-content")
        storage_id = resp["identifier"]

        artifact.checksum.value = checksum
        artifact.storage_id.value = storage_id
        artifact.status.value = ArtifactStatus.READY.value
        if artifact.name.value != definition.artifact_name.value:
            artifact.name.value = definition.artifact_name.value
        await artifact.save()

        return ArtifactGenerateResult(changed=True, checksum=checksum, storage_id=storage_id, artifact_id=artifact.id)

    async def render_artifact(
        self,
        artifact: CoreArtifact,
        artifact_created: bool,
        message: CheckArtifactCreate | RequestArtifactGenerate,
    ) -> ArtifactGenerateResult:
        response = await self.sdk.query_gql_query(
            name=message.query_id,
            variables=message.variables,
            update_group=True,
            subscribers=[artifact.id],
            tracker="artifact-query-graphql-data",
            branch_name=message.branch_name,
            timeout=message.timeout,
        )
        branch = registry.get_branch_from_registry(branch=message.branch_name)

        previous_checksum = artifact.checksum.value
        previous_storage_id = artifact.storage_id.value

        if message.transform_type == InfrahubKind.TRANSFORMJINJA2:
            artifact_content = await self.render_jinja2_template.with_options(timeout_seconds=message.timeout)(
                commit=message.commit, location=message.transform_location, data=response
            )  # type: ignore[misc]
        elif message.transform_type == InfrahubKind.TRANSFORMPYTHON:
            artifact_content = await self.execute_python_transform.with_options(timeout_seconds=message.timeout)(
                client=self.sdk,
                branch_name=message.branch_name,
                commit=message.commit,
                location=message.transform_location,
                data=response,
                convert_query_response=message.convert_query_response,
            )  # type: ignore[misc]

        if message.content_type == ContentType.APPLICATION_JSON.value and isinstance(artifact_content, dict):
            artifact_content_str = ujson.dumps(artifact_content, indent=2)
        elif message.content_type == ContentType.APPLICATION_YAML.value and isinstance(artifact_content, dict):
            artifact_content_str = yaml.dump(artifact_content, indent=2)
        else:
            artifact_content_str = str(artifact_content)

        checksum = hashlib.md5(bytes(artifact_content_str, encoding="utf-8"), usedforsecurity=False).hexdigest()

        if artifact.checksum.value == checksum:
            return ArtifactGenerateResult(
                changed=False, checksum=checksum, storage_id=artifact.storage_id.value, artifact_id=artifact.id
            )

        resp = await self.sdk.object_store.upload(content=artifact_content_str, tracker="artifact-upload-content")
        storage_id = resp["identifier"]

        artifact.content_type.value = message.content_type
        artifact.checksum.value = checksum
        artifact.storage_id.value = storage_id
        artifact.status.value = ArtifactStatus.READY.value
        if artifact.name.value != message.artifact_name:
            artifact.name.value = message.artifact_name
        await artifact.save()

        event_class = ArtifactCreatedEvent if artifact_created else ArtifactUpdatedEvent

        event = event_class(
            node_id=artifact.id,
            target_id=message.target_id,
            target_kind=message.target_kind,
            artifact_definition_id=message.artifact_definition,
            artifact_definition_name=message.artifact_definition_name,
            meta=EventMeta.from_context(context=message.context, branch=branch),
            checksum=checksum,
            checksum_previous=previous_checksum,
            storage_id=storage_id,
            storage_id_previous=previous_storage_id,
        )

        event_service = await get_event_service()
        await event_service.send(event=event)
        return ArtifactGenerateResult(changed=True, checksum=checksum, storage_id=storage_id, artifact_id=artifact.id)


def repo_object_type_to_file_type(repo_object: RepositoryObjects) -> type[InfrahubFile]:
    match repo_object:
        case RepositoryObjects.OBJECT:
            return ObjectFile
        case RepositoryObjects.MENU:
            return MenuFile
        case _:
            raise ValueError(f"Unknown repository object type: {repo_object}")



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, Self, cast

import httpx
from graphene import Boolean, Field, InputObjectType, Mutation, String

from infrahub import config
from infrahub.core.constants import InfrahubKind
from infrahub.core.manager import NodeManager
from infrahub.core.protocols import CoreReadOnlyRepository, CoreRepository
from infrahub.core.schema import NodeSchema
from infrahub.git.models import (
    GitRepositoryImportObjects,
    GitRepositoryPullReadOnly,
)
from infrahub.graphql.types.common import IdentifierInput
from infrahub.log import get_logger
from infrahub.message_bus import messages
from infrahub.message_bus.messages.git_repository_connectivity import GitRepositoryConnectivityResponse
from infrahub.repositories.create_repository import RepositoryFinalizer
from infrahub.workflows.catalogue import (
    GIT_REPOSITORIES_IMPORT_OBJECTS,
    GIT_REPOSITORIES_PULL_READ_ONLY,
)

from ...core.node.create import create_node
from ..types.task import TaskInfo
from .main import InfrahubMutationMixin, InfrahubMutationOptions, build_graphql_response

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.database import InfrahubDatabase
    from infrahub.graphql.initialization import GraphqlContext

log = get_logger()


class InfrahubRepositoryMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(cls, schema: NodeSchema | None = None, _meta=None, **options) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)

        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
        override_data: dict[str, Any] | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context
        cleanup_payload(data)
        db = database or graphql_context.db
        create_data = dict(data)
        create_data.update(override_data or {})
        obj = await create_node(data=create_data, db=db, branch=branch, schema=cls._meta.active_schema)

        await RepositoryFinalizer(
            account_session=graphql_context.active_account_session,
            services=graphql_context.active_service,
            context=graphql_context.get_context(),
        ).post_create(
            obj=obj,  # type: ignore
            branch=branch,
            db=db,
        )

        graphql_response = await build_graphql_response(info=info, db=db, obj=obj)
        return obj, cls(**graphql_response)

    @classmethod
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,  # noqa: ARG003
        node: Node | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context

        cleanup_payload(data)
        if not node:
            node: CoreReadOnlyRepository | CoreRepository = await NodeManager.get_one_by_id_or_default_filter(
                db=graphql_context.db,
                kind=cls._meta.schema.kind,
                id=data.get("id"),
                branch=branch,
                include_owner=True,
                include_source=True,
            )
        if node.get_kind() != InfrahubKind.READONLYREPOSITORY:
            return await super().mutate_update(info, data, branch, database=graphql_context.db, node=node)

        node = cast(CoreReadOnlyRepository, node)
        current_commit = node.commit.value
        current_ref = node.ref.value
        new_commit = None
        if data.commit and data.commit.value:
            new_commit = data.commit.value
        new_ref = None
        if data.ref and data.ref.value:
            new_ref = data.ref.value

        obj, result = await super().mutate_update(info, data, branch, database=graphql_context.db, node=node)
        obj = cast(CoreReadOnlyRepository, obj)

        send_update_message = (new_commit and new_commit != current_commit) or (new_ref and new_ref != current_ref)
        if not send_update_message:
            return obj, result

        log.info(
            "update read-only repository commit",
            name=obj.name.value,
            commit=data.commit.value if data.commit else None,
            ref=data.ref.value if data.ref else None,
        )

        model = GitRepositoryPullReadOnly(
            repository_id=obj.id,
            repository_name=obj.name.value,
            location=obj.location.value,
            ref=obj.ref.value,
            commit=new_commit,
            infrahub_branch_name=branch.name,
            infrahub_branch_id=str(branch.get_uuid()),
        )
        if graphql_context.service:
            await graphql_context.service.workflow.submit_workflow(
                workflow=GIT_REPOSITORIES_PULL_READ_ONLY,
                context=graphql_context.get_context(),
                parameters={"model": model},
            )
        return obj, result


def cleanup_payload(data: InputObjectType | dict[str, Any]) -> None:
    """If the input payload contains an http URL that doesn't end in .git it will be added to the payload"""
    http_without_dotgit = r"^(https?://)(?!.*\.git$).*"
    if (
        data.get("location")
        and data["location"].get("value")
        and re.match(http_without_dotgit, data["location"]["value"])
    ):
        url = httpx.URL(data["location"]["value"])
        if url.host in config.SETTINGS.git.append_git_suffix:
            data["location"]["value"] = f"{data['location']['value']}.git"


class ProcessRepository(Mutation):
    class Arguments:
        data = IdentifierInput(required=True)

    ok = Boolean()
    task = Field(TaskInfo, required=False)

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: IdentifierInput,
    ) -> dict[str, bool]:
        graphql_context: GraphqlContext = info.context
        branch = graphql_context.branch
        repository_id = str(data.id)
        repo: CoreReadOnlyRepository | CoreRepository = await NodeManager.get_one_by_id_or_default_filter(
            db=graphql_context.db,
            kind=InfrahubKind.GENERICREPOSITORY,
            id=str(data.id),
            branch=branch,
        )

        model = GitRepositoryImportObjects(
            repository_id=repository_id,
            repository_name=str(repo.name.value),
            repository_kind=repo.get_kind(),
            commit=str(repo.commit.value),
            infrahub_branch_name=branch.name,
        )
        workflow = await graphql_context.active_service.workflow.submit_workflow(
            workflow=GIT_REPOSITORIES_IMPORT_OBJECTS, context=graphql_context.get_context(), parameters={"model": model}
        )
        task = {"id": workflow.id}
        return cls(ok=True, task=task)


class ValidateRepositoryConnectivity(Mutation):
    class Arguments:
        data = IdentifierInput(required=True)

    ok = Boolean(required=True)
    message = String(required=True)

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: IdentifierInput,
    ) -> dict[str, Any]:
        graphql_context: GraphqlContext = info.context
        branch = graphql_context.branch
        repository_id = str(data.id)
        repo: CoreReadOnlyRepository | CoreRepository = await NodeManager.get_one_by_id_or_default_filter(
            db=graphql_context.db,
            kind=InfrahubKind.GENERICREPOSITORY,
            id=repository_id,
            branch=branch,
        )

        message = messages.GitRepositoryConnectivity(
            repository_name=str(repo.name.value),
            repository_location=str(repo.location.value),
        )
        if graphql_context.service:
            response = await graphql_context.service.message_bus.rpc(
                message=message, response_class=GitRepositoryConnectivityResponse
            )

        return {"ok": response.data.success, "message": response.data.message}



from collections import defaultdict
from dataclasses import asdict
from typing import AsyncGenerator

from infrahub.core.query import QueryType
from infrahub.database import InfrahubDatabase

from .models import EdgeToAdd


class PatchPlanEdgeAdder:
    def __init__(self, db: InfrahubDatabase, batch_size_limit: int = 1000) -> None:
        self.db = db
        self.batch_size_limit = batch_size_limit

    async def _run_add_query(self, edge_type: str, edges_to_add: list[EdgeToAdd]) -> dict[str, str]:
        query = """
UNWIND $edges_to_add AS edge_to_add
MATCH (a) WHERE %(id_func_name)s(a) = edge_to_add.from_id
MATCH (b) WHERE %(id_func_name)s(b) = edge_to_add.to_id
CREATE (a)-[e:%(edge_type)s]->(b)
SET e = edge_to_add.after_props
RETURN edge_to_add.identifier AS abstract_id, %(id_func_name)s(e) AS db_id
        """ % {
            "edge_type": edge_type,
            "id_func_name": self.db.get_id_function_name(),
        }
        edges_to_add_dicts = [asdict(v) for v in edges_to_add]
        # use transaction to make sure we record the results before committing them
        try:
            txn_db = self.db.start_transaction()
            async with txn_db as txn:
                results = await txn.execute_query(
                    query=query, params={"edges_to_add": edges_to_add_dicts}, type=QueryType.WRITE
                )
            abstract_to_concrete_id_map: dict[str, str] = {}
            for result in results:
                abstract_id = result.get("abstract_id")
                concrete_id = result.get("db_id")
                abstract_to_concrete_id_map[abstract_id] = concrete_id
        finally:
            await txn_db.close()
        return abstract_to_concrete_id_map

    async def execute(
        self,
        edges_to_add: list[EdgeToAdd],
    ) -> AsyncGenerator[dict[str, str], None]:
        """
        Create edges_to_add on the database.
        Returns a generator that yields dictionaries mapping EdgeToAdd.identifier to the database-level ID of the newly created edge.
        """
        edges_map_queue: dict[str, list[EdgeToAdd]] = defaultdict(list)
        for edge_to_add in edges_to_add:
            edges_map_queue[edge_to_add.edge_type].append(edge_to_add)
            if len(edges_map_queue[edge_to_add.edge_type]) > self.batch_size_limit:
                yield await self._run_add_query(
                    edge_type=edge_to_add.edge_type,
                    edges_to_add=edges_map_queue[edge_to_add.edge_type],
                )
                edges_map_queue[edge_to_add.edge_type] = []

        for edge_type, edges_group in edges_map_queue.items():
            yield await self._run_add_query(edge_type=edge_type, edges_to_add=edges_group)



from __future__ import annotations

from dataclasses import dataclass, field

from prefect import task
from prefect.cache_policies import NONE
from prefect.logging import get_run_logger

from infrahub.core.registry import registry
from infrahub.database import InfrahubDatabase  # noqa: TC001  needed for prefect flow

from .models import DisplayLabelTriggerDefinition


@dataclass
class BranchScope:
    name: str
    out_of_scope: list[str] = field(default_factory=list)


@task(
    name="gather-trigger-display-labels-jinja2",
    cache_policy=NONE,
)
async def gather_trigger_display_labels_jinja2(
    db: InfrahubDatabase | None = None,  # noqa: ARG001 Needed to have a common function signature for gathering functions
) -> list[DisplayLabelTriggerDefinition]:
    log = get_run_logger()

    # Build a list of all branches to process based on which branch is different from main
    branches_with_diff_from_main = registry.get_altered_schema_branches()
    branches_to_process: list[BranchScope] = [BranchScope(name=branch) for branch in branches_with_diff_from_main]
    branches_to_process.append(BranchScope(name=registry.default_branch, out_of_scope=branches_with_diff_from_main))

    triggers: list[DisplayLabelTriggerDefinition] = []

    for branch in branches_to_process:
        schema_branch = registry.schema.get_schema_branch(name=branch.name)
        branch_triggers = DisplayLabelTriggerDefinition.from_schema_display_labels(
            branch=branch.name,
            display_labels=schema_branch.display_labels,
            branches_out_of_scope=branch.out_of_scope,
        )
        log.info(f"Generating {len(branch_triggers)} Jinja2 trigger for {branch.name} (except {branch.out_of_scope})")

        triggers.extend(branch_triggers)

    return triggers



from typing import Any

from fastapi.middleware.gzip import GZipMiddleware
from starlette.middleware.cors import CORSMiddleware
from starlette.types import ASGIApp, Receive, Scope, Send

from infrahub import config


class InfrahubCORSMiddleware(CORSMiddleware):
    def __init__(self, app: ASGIApp, *args: Any, **kwargs: Any) -> None:
        config.SETTINGS.initialize_and_exit()
        kwargs["allow_origins"] = config.SETTINGS.api.cors_allow_origins
        kwargs["allow_credentials"] = config.SETTINGS.api.cors_allow_credentials
        kwargs["allow_methods"] = config.SETTINGS.api.cors_allow_methods
        kwargs["allow_headers"] = config.SETTINGS.api.cors_allow_headers

        super().__init__(app, *args, **kwargs)


class ConditionalGZipMiddleware(GZipMiddleware):
    def __init__(
        self,
        app: ASGIApp,
        *,
        minimum_size: int = 500,
        compresslevel: int = 9,
        include_paths: tuple[str, ...] = (),
    ) -> None:
        super().__init__(app, minimum_size=minimum_size, compresslevel=compresslevel)
        self.include_paths = include_paths

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:  # type: ignore[override]
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        path = scope.get("path", "")
        if any(path.startswith(include) for include in self.include_paths):
            await super().__call__(scope, receive, send)
        else:
            await self.app(scope, receive, send)



from __future__ import annotations

import asyncio
import time
import uuid
from asyncio import Lock as LocalLock
from asyncio import sleep
from contextvars import ContextVar
from typing import TYPE_CHECKING

import redis.asyncio as redis
from prometheus_client import Histogram
from redis.asyncio.lock import Lock as GlobalLock

from infrahub import config
from infrahub.core.timestamp import current_timestamp
from infrahub.worker import WORKER_IDENTITY

if TYPE_CHECKING:
    from types import TracebackType

    from infrahub.services import InfrahubServices


registry: InfrahubLockRegistry = None


METRIC_PREFIX = "infrahub_lock"
LOCK_PREFIX = "lock"

LOCK_ACQUIRE_TIME_METRICS = Histogram(
    f"{METRIC_PREFIX}_acquire_seconds",
    "Time to acquire the lock on a given object",
    labelnames=["lock", "type"],
    buckets=[0.0005, 0.25, 0.5, 1, 5],
)

LOCK_RESERVE_TIME_METRICS = Histogram(
    f"{METRIC_PREFIX}_reserved_duration_seconds",
    "Time while a given lock is reserved by a given client",
    labelnames=["lock", "type"],
    buckets=[0.001, 0.5, 1, 5, 10],
)

LOCAL_SCHEMA_LOCK = "local.schema"
GLOBAL_INIT_LOCK = "global.init"
GLOBAL_SCHEMA_LOCK = "global.schema"
GLOBAL_GRAPH_LOCK = "global.graph"


class InfrahubMultiLock:
    """Context manager to allow multiple locks to be reserved together"""

    def __init__(
        self, lock_registry: InfrahubLockRegistry, locks: list[str] | None = None, metrics: bool = True
    ) -> None:
        self.registry = lock_registry
        self.locks = locks or []
        self.metrics = metrics

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ):
        await self.release()

    async def acquire(self) -> None:
        for lock in self.locks:
            await self.registry.get(name=lock, metrics=self.metrics).acquire()

    async def release(self) -> None:
        for lock in reversed(self.locks):
            await self.registry.get(name=lock, metrics=self.metrics).release()


class NATSLock:
    """Context manager to lock using NATS"""

    def __init__(self, service: InfrahubServices, name: str) -> None:
        self.name = name
        self.token = None
        self.service = service

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ):
        await self.release()

    async def acquire(self) -> None:
        token = f"{current_timestamp()}::{WORKER_IDENTITY}"
        while True:
            if await self.do_acquire(token):
                self.token = token
                return
            await sleep(0.1)  # default Redis GlobalLock value

    async def do_acquire(self, token: str) -> bool | None:
        return await self.service.cache.set(key=self.name, value=token, not_exists=True)

    async def release(self) -> None:
        await self.service.cache.delete(key=self.name)

    async def locked(self) -> bool:
        return await self.service.cache.get(key=self.name) is not None


class InfrahubLock:
    """InfrahubLock object to provide a unified interface for both Local and Distributed locks.

    Having the same interface for both local and distributed tests will simplify our unit tests.
    """

    def __init__(
        self,
        name: str,
        connection: redis.Redis | InfrahubServices | None = None,
        local: bool | None = None,
        in_multi: bool = False,
        metrics: bool = True,
    ) -> None:
        self.use_local: bool | None = local
        self.local: LocalLock = None
        self.remote: GlobalLock = None
        self.name: str = name
        self.connection: redis.Redis | None = connection
        self.in_multi: bool = in_multi
        self.lock_type: str = "multi" if self.in_multi else "individual"
        self._acquire_time: int | None = None
        self.event = asyncio.Event()
        self._recursion_var: ContextVar[int | None] = ContextVar(f"infrahub_lock_recursion_{self.name}", default=None)
        self.metrics = metrics

        if not self.connection or (self.use_local is None and name.startswith("local.")):
            self.use_local = True

        if self.use_local:
            self.local = LocalLock()
        elif config.SETTINGS.cache.driver == config.CacheDriver.Redis:
            self.remote = GlobalLock(redis=self.connection, name=f"{LOCK_PREFIX}.{self.name}")
        else:
            self.remote = NATSLock(service=self.connection, name=f"{LOCK_PREFIX}.{self.name}")

    @property
    def acquire_time(self) -> int:
        if self._acquire_time is not None:
            return self._acquire_time

        raise ValueError("The lock has not been initialized")

    @acquire_time.setter
    def acquire_time(self, value: int) -> None:
        self._acquire_time = value

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ):
        await self.release()

    async def acquire(self) -> None:
        depth = self._recursion_var.get()
        if depth is not None:
            self._recursion_var.set(depth + 1)
            return

        if self.metrics:
            with LOCK_ACQUIRE_TIME_METRICS.labels(self.name, self.lock_type).time():
                if not self.use_local:
                    await self.remote.acquire(token=f"{current_timestamp()}::{WORKER_IDENTITY}")
                else:
                    await self.local.acquire()
        elif not self.use_local:
            await self.remote.acquire(token=f"{current_timestamp()}::{WORKER_IDENTITY}")
        else:
            await self.local.acquire()

        self.acquire_time = time.time_ns()
        self.event.clear()
        self._recursion_var.set(1)

    async def release(self) -> None:
        depth = self._recursion_var.get()
        if depth is None:
            raise RuntimeError("Lock release attempted without ownership context.")

        if depth > 1:
            self._recursion_var.set(depth - 1)
            return

        if self.acquire_time is not None:
            duration_ns = time.time_ns() - self.acquire_time
            if self.metrics:
                LOCK_RESERVE_TIME_METRICS.labels(self.name, self.lock_type).observe(duration_ns / 1000000000)
            self.acquire_time = None

        if not self.use_local:
            await self.remote.release()
        else:
            self.local.release()

        self._recursion_var.set(None)
        self.event.set()

    async def locked(self) -> bool:
        if not self.use_local:
            return await self.remote.locked()

        return self.local.locked()


class LockNameGenerator:
    local = "local"
    _global = "global"

    def generate_name(self, name: str, namespace: str | None = None, local: bool | None = None) -> str:
        if namespace is None and local is None:
            return name

        new_name = ""
        if local is True:
            new_name = f"{self.local}."
        elif local is False:
            new_name = f"{self._global}."

        if namespace is not None:
            new_name += f"{namespace}."
        new_name += name

        return new_name

    def unpack_name(self, name: str) -> tuple[str, str | None, bool | None]:
        local = None
        namespace = None

        parts = name.split(".")
        if parts[0] == self.local:
            local = True
            parts = parts[1:]
        elif parts[0] == self._global:
            local = False
            parts = parts[1:]

        if len(parts) > 1:
            namespace = parts[0]
            original_name = ".".join(parts[1:])
        else:
            original_name = parts[0]

        return original_name, namespace, local


class InfrahubLockRegistry:
    def __init__(
        self,
        token: str | None = None,
        local_only: bool = False,
        service: InfrahubServices | None = None,
        name_generator: LockNameGenerator | None = None,
    ) -> None:
        if config.SETTINGS.cache.enable and not local_only:
            if config.SETTINGS.cache.driver == config.CacheDriver.Redis:
                self.connection = redis.Redis(
                    host=config.SETTINGS.cache.address,
                    port=config.SETTINGS.cache.service_port,
                    db=config.SETTINGS.cache.database,
                    ssl=config.SETTINGS.cache.tls_enabled,
                    ssl_cert_reqs="optional" if not config.SETTINGS.cache.tls_insecure else "none",
                    ssl_check_hostname=not config.SETTINGS.cache.tls_insecure,
                    ssl_ca_certs=config.SETTINGS.cache.tls_ca_file,
                )
            else:
                self.connection = service
        else:
            self.connection = None

        self.token = token or str(uuid.uuid4())
        self.locks: dict[str, InfrahubLock] = {}
        self.name_generator = name_generator or LockNameGenerator()

    def get_existing(
        self,
        name: str,
        namespace: str | None,
        local: bool | None = None,
    ) -> InfrahubLock | None:
        lock_name = self.name_generator.generate_name(name=name, namespace=namespace, local=local)
        if lock_name not in self.locks:
            return None
        return self.locks[lock_name]

    def get(
        self,
        name: str,
        namespace: str | None = None,
        local: bool | None = None,
        in_multi: bool = False,
        metrics: bool = True,
    ) -> InfrahubLock:
        lock_name = self.name_generator.generate_name(name=name, namespace=namespace, local=local)
        if lock_name not in self.locks:
            self.locks[lock_name] = InfrahubLock(
                name=lock_name, connection=self.connection, in_multi=in_multi, metrics=metrics
            )
        return self.locks[lock_name]

    def local_schema_lock(self) -> LocalLock:
        return self.get(name=LOCAL_SCHEMA_LOCK)

    def initialization(self) -> LocalLock:
        return self.get(name=GLOBAL_INIT_LOCK)

    async def local_schema_wait(self) -> None:
        await self.get(name=LOCAL_SCHEMA_LOCK).event.wait()

    def global_schema_lock(self) -> InfrahubMultiLock:
        return InfrahubMultiLock(lock_registry=self, locks=[LOCAL_SCHEMA_LOCK, GLOBAL_SCHEMA_LOCK])

    def global_graph_lock(self) -> InfrahubMultiLock:
        return InfrahubMultiLock(lock_registry=self, locks=[LOCAL_SCHEMA_LOCK, GLOBAL_GRAPH_LOCK, GLOBAL_SCHEMA_LOCK])


def initialize_lock(local_only: bool = False, service: InfrahubServices | None = None) -> None:
    global registry
    registry = InfrahubLockRegistry(local_only=local_only, service=service)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.constants import RelationshipStatus
from infrahub.core.graph.schema import GraphNodeRelationships, GraphRelDirection
from infrahub.core.query import Query, QueryType
from infrahub.core.timestamp import Timestamp

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class DeleteElementInSchemaQuery(Query):
    name = "delete_element_in_schema"
    type = QueryType.WRITE
    insert_return: bool = False

    def __init__(
        self,
        element_names: list[str],
        node_name: str,
        node_namespace: str,
        **kwargs: Any,
    ) -> None:
        self.element_names = element_names
        self.node_name = node_name
        self.node_namespace = node_namespace

        super().__init__(**kwargs)

    def render_match(self) -> str:
        query = """
        MATCH path = (attr_node:Node)-[:HAS_ATTRIBUTE]->(attr:Attribute)
        MATCH (attr_node)-[:HAS_ATTRIBUTE]->(attr_name:Attribute)-[:HAS_VALUE]->(attr_value:AttributeValue)
        """
        return query

    def render_where(self) -> str:
        at = self.at or Timestamp()
        filters, params = at.get_query_filter_path()
        self.params.update(params)

        # ruff: noqa: E501
        query = """
        WHERE ( "SchemaAttribute" in LABELS(attr_node) OR "SchemaRelationship" IN LABELS(attr_node))
            AND exists( (attr_node)-[:IS_RELATED]->(:Relationship)<-[:IS_RELATED]-(:Node)-[:HAS_ATTRIBUTE]->(:Attribute { name: "name"})-[:HAS_VALUE]->(:AttributeValue { value: $node_name }) )
            AND exists( (attr_node)-[:IS_RELATED]->(:Relationship)<-[:IS_RELATED]-(:Node)-[:HAS_ATTRIBUTE]->(:Attribute { name: "namespace"})-[:HAS_VALUE]->(:AttributeValue  { value: $node_namespace }) )
            AND ( attr_name.name = "name" AND attr_value.value IN $element_names)
            AND all(r IN relationships(path) WHERE ( %(filters)s ))
        """ % {"filters": filters}

        return query

    @staticmethod
    def _render_sub_query_per_rel_type(rel_name: str, rel_type: str, direction: GraphRelDirection) -> str:
        subquery = [
            f"WITH peer_node, {rel_name}, element_to_delete",
            f'WHERE type({rel_name}) = "{rel_type}"',
        ]
        if direction == GraphRelDirection.OUTBOUND:
            subquery.append(f"CREATE (element_to_delete)-[:{rel_type} $rel_props_prev ]->(peer_node)")
        elif direction == GraphRelDirection.INBOUND:
            subquery.append(f"CREATE (element_to_delete)<-[:{rel_type} $rel_props_prev ]-(peer_node)")
        subquery.append("RETURN peer_node as p2")
        return "\n".join(subquery)

    @classmethod
    def _render_sub_query_out(cls) -> tuple[str, str]:
        rel_name = "rel_outband"
        sub_query_out_args = f"peer_node, {rel_name}, element_to_delete"
        sub_queries_out = [
            cls._render_sub_query_per_rel_type(
                rel_name=rel_name, rel_type=rel_type, direction=GraphRelDirection.OUTBOUND
            )
            for rel_type, rel_def in GraphNodeRelationships.model_fields.items()
            if rel_def.default.direction in [GraphRelDirection.OUTBOUND, GraphRelDirection.EITHER]
        ]
        sub_query_out = "\nUNION\n".join(sub_queries_out)
        return sub_query_out, sub_query_out_args

    @classmethod
    def _render_sub_query_in(cls) -> tuple[str, str]:
        rel_name = "rel_inband"
        sub_query_in_args = f"peer_node, {rel_name}, element_to_delete"
        sub_queries_in = [
            cls._render_sub_query_per_rel_type(
                rel_name=rel_name, rel_type=rel_type, direction=GraphRelDirection.INBOUND
            )
            for rel_type, rel_def in GraphNodeRelationships.model_fields.items()
            if rel_def.default.direction in [GraphRelDirection.INBOUND, GraphRelDirection.EITHER]
        ]
        sub_query_in = "\nUNION\n".join(sub_queries_in)
        return sub_query_in, sub_query_in_args

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string())
        self.params.update(branch_params)

        self.params["element_names"] = self.element_names
        self.params["node_name"] = self.node_name
        self.params["node_namespace"] = self.node_namespace

        self.params["current_time"] = self.at.to_string()
        self.params["branch_name"] = self.branch.name

        self.params["rel_props_prev"] = {
            "branch": self.branch.name,
            "branch_level": self.branch.hierarchy_level,
            "status": RelationshipStatus.DELETED.value,
            "from": self.at.to_string(),
        }

        sub_query_out, sub_query_out_args = self._render_sub_query_out()
        sub_query_in, sub_query_in_args = self._render_sub_query_in()

        self.add_to_query(self.render_match())
        self.add_to_query(self.render_where())

        # ruff: noqa: E501
        query = """
        CALL (attr_node) {
            MATCH (root:Root)<-[r:IS_PART_OF]-(attr_node)
            WHERE %(branch_filter)s
            RETURN attr_node as n1, r as r1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as element_to_delete, r1 as rb
        WHERE rb.status = "active"
        WITH DISTINCT(element_to_delete) as element_to_delete

        // Process Outbound Relationship
        MATCH (element_to_delete)-[]->(peer)
        CALL (element_to_delete, peer) {
            MATCH (element_to_delete)-[r]->(peer)
            WHERE %(branch_filter)s
            RETURN element_to_delete as n1, r as rel_outband1, peer as p1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as element_to_delete, rel_outband1 as rel_outband, p1 as peer_node
        WHERE rel_outband.status = "active"
        CALL (%(sub_query_out_args)s) {
            %(sub_query_out)s
        }
        WITH p2 as peer_node, rel_outband, element_to_delete
        FOREACH (i in CASE WHEN rel_outband.branch = $branch_name THEN [1] ELSE [] END |
            SET rel_outband.to = $current_time
        )
        WITH DISTINCT(element_to_delete) AS element_to_delete
        // Process Inbound Relationship
        MATCH (element_to_delete)<-[]-(peer)
        CALL (element_to_delete, peer) {
            MATCH (element_to_delete)<-[r]-(peer)
            WHERE %(branch_filter)s
            RETURN element_to_delete as n1, r as rel_inband1, peer as p1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as element_to_delete, rel_inband1 as rel_inband, p1 as peer_node
        WHERE rel_inband.status = "active"
        CALL (%(sub_query_in_args)s) {
            %(sub_query_in)s
        }
        WITH p2 as peer_node, rel_inband, element_to_delete
        FOREACH (i in CASE WHEN rel_inband.branch = $branch_name THEN [1] ELSE [] END |
            SET rel_inband.to = $current_time
        )
        RETURN DISTINCT element_to_delete
        """ % {
            "branch_filter": branch_filter,
            "sub_query_out": sub_query_out,
            "sub_query_in": sub_query_in,
            "sub_query_out_args": sub_query_out_args,
            "sub_query_in_args": sub_query_in_args,
        }
        self.add_to_query(query)



from __future__ import annotations

from typing import TYPE_CHECKING, Self

from graphene import Boolean, Field, InputObjectType, Mutation, String

from infrahub import lock
from infrahub.core import registry
from infrahub.core.constants import RESTRICTED_NAMESPACES, GlobalPermissions
from infrahub.core.manager import NodeManager
from infrahub.core.schema import DropdownChoice, GenericSchema, NodeSchema
from infrahub.database import InfrahubDatabase, retry_db_transaction
from infrahub.events import EventMeta
from infrahub.events.schema_action import SchemaUpdatedEvent
from infrahub.exceptions import ValidationError
from infrahub.graphql.context import apply_external_context
from infrahub.graphql.types.context import ContextInput
from infrahub.log import get_log_data, get_logger
from infrahub.permissions import define_global_permission_from_branch
from infrahub.worker import WORKER_IDENTITY

from ..types import DropdownFields

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.context import InfrahubContext
    from infrahub.core.branch import Branch
    from infrahub.services import InfrahubServices

    from ..initialization import GraphqlContext

log = get_logger()


def _validate_schema_permission(graphql_context: GraphqlContext) -> None:
    graphql_context.active_permissions.raise_for_permission(
        permission=define_global_permission_from_branch(
            permission=GlobalPermissions.MANAGE_SCHEMA, branch_name=graphql_context.branch.name
        )
    )


class SchemaEnumInput(InputObjectType):
    kind = String(required=True)
    attribute = String(required=True)
    enum = String(required=True)


class SchemaDropdownRemoveInput(InputObjectType):
    kind = String(required=True)
    attribute = String(required=True)
    dropdown = String(required=True)


class SchemaDropdownAddInput(SchemaDropdownRemoveInput):
    color = String(required=False)
    description = String(required=False)
    label = String(required=False)


class SchemaDropdownAdd(Mutation):
    class Arguments:
        data = SchemaDropdownAddInput(required=True)
        context = ContextInput(required=False)

    ok = Boolean()
    object = Field(DropdownFields)

    @classmethod
    @retry_db_transaction(name="schema_dropdown_add")
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: SchemaDropdownAddInput,
        context: ContextInput | None = None,
    ) -> Self:
        graphql_context: GraphqlContext = info.context

        _validate_schema_permission(graphql_context=graphql_context)
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        kind = graphql_context.db.schema.get(name=str(data.kind), branch=graphql_context.branch.name, duplicate=False)
        attribute = str(data.attribute)
        validate_kind_dropdown(kind=kind, attribute=attribute)
        dropdown = str(data.dropdown)
        choice = DropdownChoice(name=dropdown, color=data.color, label=data.label, description=data.description)

        if found_attribute := [attrib for attrib in kind.attributes if attrib.name == attribute]:
            attrib = found_attribute[0]
            if [dropdown_entry for dropdown_entry in attrib.choices if dropdown_entry.name == dropdown]:
                raise ValidationError(
                    f"The dropdown value {dropdown} already exists on {kind.kind} in attribute {attribute}"
                )
            attrib.choices.append(choice)

        await update_registry(
            kind=kind,
            branch=graphql_context.branch,
            db=graphql_context.db,
            account_id=graphql_context.active_account_session.account_id,
            service=graphql_context.active_service,
            context=graphql_context.get_context(),
        )

        kind = graphql_context.db.schema.get(name=str(data.kind), branch=graphql_context.branch.name, duplicate=False)
        attrib = kind.get_attribute(attribute)
        dropdown_entry = {}
        success = False

        for entry in attrib.choices:
            if entry.name == dropdown:
                dropdown_entry = {
                    "value": dropdown,
                    "color": entry.color,
                    "label": entry.label,
                    "description": entry.description,
                }
                success = True

        return cls(object=dropdown_entry, ok=success)


class SchemaDropdownRemove(Mutation):
    class Arguments:
        data = SchemaDropdownRemoveInput(required=True)
        context = ContextInput(required=False)

    ok = Boolean()

    @classmethod
    @retry_db_transaction(name="schema_dropdown_remove")
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: SchemaDropdownRemoveInput,
        context: ContextInput | None = None,
    ) -> dict[str, bool]:
        graphql_context: GraphqlContext = info.context

        _validate_schema_permission(graphql_context=graphql_context)
        kind = graphql_context.db.schema.get(name=str(data.kind), branch=graphql_context.branch.name, duplicate=False)
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        attribute = str(data.attribute)
        validate_kind_dropdown(kind=kind, attribute=attribute)
        dropdown = str(data.dropdown)
        nodes_with_dropdown = await NodeManager.query(
            db=graphql_context.db,
            schema=kind.kind,
            filters={f"{attribute}__value": dropdown},
            branch=graphql_context.branch,
        )
        if nodes_with_dropdown:
            raise ValidationError(f"There are still {kind.kind} objects using this dropdown")

        if found_attribute := [attrib for attrib in kind.attributes if attrib.name == attribute]:
            attrib = found_attribute[0]
            if not [dropdown_entry for dropdown_entry in attrib.choices if dropdown_entry.name == dropdown]:
                raise ValidationError(
                    f"The dropdown value {dropdown} does not exists on {kind.kind} in attribute {attribute}"
                )
            if len(attrib.choices) == 1:
                raise ValidationError(f"Unable to remove the last dropdown on {kind.kind} in attribute {attribute}")
            attrib.choices = [entry for entry in attrib.choices if dropdown != entry.name]

        await update_registry(
            kind=kind,
            branch=graphql_context.branch,
            db=graphql_context.db,
            account_id=graphql_context.active_account_session.account_id,
            service=graphql_context.active_service,
            context=graphql_context.get_context(),
        )

        return {"ok": True}


class SchemaEnumAdd(Mutation):
    class Arguments:
        data = SchemaEnumInput(required=True)
        context = ContextInput(required=False)

    ok = Boolean()

    @classmethod
    @retry_db_transaction(name="schema_dropdown_add")
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: SchemaEnumInput,
        context: ContextInput | None = None,
    ) -> dict[str, bool]:
        graphql_context: GraphqlContext = info.context

        _validate_schema_permission(graphql_context=graphql_context)
        kind = graphql_context.db.schema.get(name=str(data.kind), branch=graphql_context.branch.name, duplicate=False)
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        attribute = str(data.attribute)
        enum = str(data.enum)
        validate_kind_enum(kind=kind, attribute=attribute)

        for attrib in kind.attributes:
            if attribute == attrib.name:
                if enum in attrib.enum:
                    raise ValidationError(
                        f"The enum value {enum} already exists on {kind.kind} in attribute {attribute}"
                    )
                attrib.enum.append(enum)

        await update_registry(
            kind=kind,
            branch=graphql_context.branch,
            db=graphql_context.db,
            account_id=graphql_context.active_account_session.account_id,
            service=graphql_context.active_service,
            context=graphql_context.get_context(),
        )

        return {"ok": True}


class SchemaEnumRemove(Mutation):
    class Arguments:
        data = SchemaEnumInput(required=True)
        context = ContextInput(required=False)

    ok = Boolean()

    @classmethod
    @retry_db_transaction(name="schema_enum_remove")
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: SchemaEnumInput,
        context: ContextInput | None = None,
    ) -> dict[str, bool]:
        graphql_context: GraphqlContext = info.context

        _validate_schema_permission(graphql_context=graphql_context)
        kind = graphql_context.db.schema.get(name=str(data.kind), branch=graphql_context.branch.name, duplicate=False)
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        attribute = str(data.attribute)
        enum = str(data.enum)
        validate_kind_enum(kind=kind, attribute=attribute)
        nodes_with_enum = await NodeManager.query(
            db=graphql_context.db,
            schema=kind.kind,
            filters={f"{attribute}__value": enum},
            branch=graphql_context.branch,
        )
        if nodes_with_enum:
            raise ValidationError(f"There are still {kind.kind} objects using this enum")

        for attrib in kind.attributes:
            if attribute == attrib.name:
                if enum not in attrib.enum:
                    raise ValidationError(
                        f"The enum value {enum} does not exists on {kind.kind} in attribute {attribute}"
                    )
                if len(attrib.enum) == 1:
                    raise ValidationError(f"Unable to remove the last enum on {kind.kind} in attribute {attribute}")
                attrib.enum = [entry for entry in attrib.enum if entry != enum]

        await update_registry(
            kind=kind,
            branch=graphql_context.branch,
            db=graphql_context.db,
            account_id=graphql_context.active_account_session.account_id,
            service=graphql_context.active_service,
            context=graphql_context.get_context(),
        )

        return {"ok": True}


def validate_kind_dropdown(kind: GenericSchema | NodeSchema, attribute: str) -> None:
    validate_kind(kind=kind, attribute=attribute)
    matching_attribute = [attrib for attrib in kind.attributes if attrib.name == attribute]
    if matching_attribute and matching_attribute[0].kind != "Dropdown":
        raise ValidationError(f"Attribute {attribute} on {kind.kind} is not a Dropdown")


def validate_kind_enum(kind: GenericSchema | NodeSchema, attribute: str) -> None:
    validate_kind(kind=kind, attribute=attribute)
    matching_attribute = [attrib for attrib in kind.attributes if attrib.name == attribute]
    if not matching_attribute[0].enum:
        raise ValidationError(f"Attribute {attribute} on {kind.kind} is not an enum")


def validate_kind(kind: GenericSchema | NodeSchema, attribute: str) -> None:
    if kind.namespace in RESTRICTED_NAMESPACES:
        raise ValidationError(f"Operation not allowed for {kind.kind} in restricted namespace {kind.namespace}")
    if attribute not in kind.attribute_names:
        raise ValidationError(f"Attribute {attribute} does not exist on {kind.kind}")

    matching_attribute = [attrib for attrib in kind.attributes if attrib.name == attribute]

    if matching_attribute[0].inherited:
        raise ValidationError(f"Attribute {attribute} on {kind.kind} is inherited and must be changed on the generic")


async def update_registry(
    kind: NodeSchema,
    db: InfrahubDatabase,
    branch: Branch,
    account_id: str,
    service: InfrahubServices,
    context: InfrahubContext,
) -> None:
    async with lock.registry.global_schema_lock():
        branch_schema = registry.schema.get_schema_branch(name=branch.name)

        # We create a copy of the existing branch schema to do some validation before loading it.
        tmp_schema = branch_schema.duplicate()

        tmp_schema.set(name=kind.kind, schema=kind)
        tmp_schema.process()

        diff = tmp_schema.diff(branch_schema)

        if diff.all:
            log.info(f"Schema has diff, will need to be updated {diff.all}", branch=branch.name)
            async with db.start_transaction() as dbt:
                await registry.schema.update_schema_branch(
                    schema=tmp_schema, db=dbt, branch=branch.name, limit=diff.all, update_db=True
                )
                branch.update_schema_hash()
                log.info("Schema has been updated", branch=branch.name, hash=branch.active_schema_hash.main)
                await branch.save(db=dbt)

            await service.component.refresh_schema_hash(branches=[branch.name])

            log_data = get_log_data()
            request_id = log_data.get("request_id", "")
            event = SchemaUpdatedEvent(
                branch_name=branch.name,
                schema_hash=branch.active_schema_hash.main,
                meta=EventMeta(
                    initiator_id=WORKER_IDENTITY,
                    request_id=request_id,
                    account_id=account_id,
                    branch=branch,
                    context=context,
                ),
            )
            await service.event.send(event=event)
